import collections

import numpy as np
import matplotlib.pyplot as plt
import scipy as sp

from varwg import helpers as my
from varwg.smoothing import smooth


def time_part_average(
    values, times, time_format_str="%w", expand=True, return_times=False
):
    """Return averages per time part, e.g. monthly means."""
    time_parts = times.time_part(times, time_format_str)
    time_parts_diffs = time_parts[1:] - time_parts[:-1]
    breaks = np.where(time_parts_diffs != 0)[0] + 1
    values_splitted = np.split(values, breaks)
    values_means = np.array([val.mean() for val in values_splitted])
    if expand:
        repeats = [len(val) for val in values_splitted]
        time_part_average.repeats = repeats
        values_means = np.repeat(values_means, repeats)
    if return_times:
        return values_means, np.r_[[times[0]], times[breaks]]
    return values_means


def matr_img(
    array,
    title=None,
    xlabels=None,
    ylabels=None,
    fig=None,
    figsize=None,
    fontsize=12,
    colorbar=False,
    text=True,
    *args,
    **kwds,
):
    fig = plt.figure(figsize=figsize) if fig is None else fig
    plt.imshow(array, interpolation="nearest", **kwds)
    if xlabels is not None:
        plt.xticks(np.arange(len(xlabels)), xlabels, rotation=45)
    if ylabels is not None:
        plt.yticks(np.arange(len(ylabels)), ylabels)
    if title is not None:
        plt.title(title)
        try:
            fig.canvas.set_window_title(
                "%s (%d)" % (title, fig.canvas.manager.num)
            )
        except AttributeError:
            pass
    ax = plt.gca()
    ax.tick_params(left="off", bottom="off", top="off", right="off")

    if colorbar:
        plt.colorbar()
    if text:
        for x in range(array.shape[0]):
            for y in range(array.shape[1]):
                # do not let yourself be fooled by the strange
                # imshow-coordinate system
                plt.text(
                    y,
                    x,
                    "%.2f" % array[x, y],
                    fontsize=fontsize,
                    horizontalalignment="center",
                    verticalalignment="center",
                )
    return fig, ax


def corr_img(data, k=0, title=None, var_names=None, rank=False, *args, **kwds):
    """Plots the correlation matrix of data using imshow"""
    if rank:
        corr = cross_rank_corr(data, k)
    else:
        corr = cross_corr(data, k)
    return matr_img(
        corr,
        title,
        xlabels=var_names,
        ylabels=var_names,
        vmin=-1,
        vmax=1,
        cmap="coolwarm",
        *args,
        **kwds,
    )


def auto_cov_(data, k):
    """Return the autocovariance-vector for lag k. Variables are assumed to
    be stored in rows, with time extending across the columns."""
    data = np.asarray(data)
    if np.ndim(data) == 1:
        data = data[None, :]
    n_vars = data.shape[0]
    k = np.atleast_1d(np.abs(k))
    finite_ii = np.isfinite(data)
    n_data = finite_ii.sum(axis=1)[None, :]
    means = np.nanmean(data, axis=1)[:, None]
    auto = np.empty((len(k), n_vars))
    for ii, lag in enumerate(k):
        if lag == 0:
            auto0 = [
                np.nanmean((row - mean) ** 2) for row, mean in zip(data, means)
            ]
            auto[ii] = np.array(auto0)
        else:
            for jj, dat in enumerate(data):
                dat1, dat2 = dat[:-lag], dat[lag:]
                lag_finite_ii = np.isfinite(dat1) & np.isfinite(dat2)
                nans = np.sum(~lag_finite_ii)
                auto[ii, jj] = np.mean(
                    (dat1[lag_finite_ii] - means)
                    * (dat2[lag_finite_ii] - means)
                )
                auto[ii, jj] *= (n_data[jj] - lag - nans) / (
                    n_data[jj]
                ).astype(float)
    return np.squeeze(auto)


def auto_cov(data, k):
    """Return the autocovariance-vector for lag k. Variables are assumed to
    be stored in rows, with time extending across the columns.

    This version uses masked arrays, but is not faster than auto_cov_, which
    has a deeper nesting for loop.
    """
    data = np.copy(data)
    if np.ndim(data) == 1:
        data = data[None, :]
    n_vars = data.shape[0]
    finite_ii = np.isfinite(data)
    n_data = finite_ii.sum(axis=1)[:, None]
    data = np.ma.array(data, mask=~finite_ii)
    data -= data.mean(axis=1)[:, None]
    k = np.atleast_1d(np.abs(k))
    auto = np.empty((len(k), n_vars), dtype=float)
    for ii, lag in enumerate(k):
        if lag == 0:
            auto0 = [(row**2).mean() for row in data]
            auto[ii] = np.array(auto0)
        else:
            data1, data2 = data[:, :-lag], data[:, lag:]
            auto[ii] = (data1 * data2).mean(axis=1)
            auto[ii] *= np.ravel(((n_data - lag) / (n_data).astype(float)))
    return np.squeeze(auto)


def auto_corr(data, k):
    """Return the autocorrelation-vector for lag k.

    Parameters
    ----------
    data : (K, T) array
        K number of variables. T time steps.
    k : int or iterable
        time-lag or time lags if k is an iterable
    """
    if not isinstance(k, collections.abc.Iterable):
        k = (k,)
    return np.squeeze([auto_cov(data, lag) / auto_cov(data, 0) for lag in k])


def _partial_autocorr_uni(data, k):
    """Univariate implementation. Call partial_autocorr for multi- and
    univariate cases instead.
    Autocorrelation with the influence of the time steps between t and t-k
    removed. See Hartung p.677f. Tested according to example on p.694"""
    k = np.atleast_1d(k)
    autos = np.empty(len(k))
    r_k = auto_corr(data, 1)
    for lag_i, lag in enumerate(k):
        if lag == 0:
            autos[lag_i] = 1
        elif lag == 1:
            autos[lag_i] = r_k
        else:
            # the following block was written for k as an integer. optimization
            # in terms of not executing stuff twice is probably possible
            # read P_k-1
            P_km1 = np.atleast_1d(auto_corr(data, list(range(1, lag))))
            P_km1_rev = P_km1[::-1]
            # read P_k-1,k-1
            P_km1km1 = np.ones((lag - 1, lag - 1))
            for ii in range(lag - 1):
                if ii > 0:
                    P_km1km1[ii, :ii] = P_km1_rev[-ii:]
                if ii < lag - 2:
                    P_km1km1[ii, ii + 1 :] = P_km1[ii:-1]
            P_km1km1_inv = np.linalg.inv(P_km1km1)
            autos[lag_i] = (
                r_k - P_km1 @ P_km1km1_inv @ P_km1_rev.T
            ) / np.sqrt(
                (1 - P_km1 @ P_km1km1_inv @ P_km1.T)
                @ (1 - P_km1_rev @ P_km1km1_inv @ P_km1_rev.T)
            )
    return np.squeeze(autos)


def partial_autocorr(data, k):
    """Autocorrelation with the influence of the time steps between t and t-k
    removed. See Hartung p.677f. Tested according to example on p.694"""
    if np.ndim(data) == 1:
        return _partial_autocorr_uni(data, k)
    else:
        return np.array([_partial_autocorr_uni(var, k) for var in data])


def plot_auto_corr(
    data,
    k_range=7 * 24,
    title="",
    var_names=None,
    n_per_day=1,
    partial=False,
    n_difference=0,
    figsize=None,
    window_title=True,
    fig=None,
    ax=None,
    *args,
    **kwds,
):
    if fig is None or ax is None:
        fig, ax = plt.subplots(
            nrows=1, ncols=1, figsize=figsize, constrained_layout=True
        )
    fig.canvas.mpl_connect("draw_event", my.scale_yticks)
    lags = np.arange(k_range)
    corr_func = partial_autocorr if partial else auto_corr
    if isinstance(data, np.ndarray):
        d_data = np.copy(data)
        if np.ndim(data) == 1:
            d_data = d_data[np.newaxis, :]
        if np.ndim(data) == 2:
            d_data = d_data[np.newaxis, ...]
    else:
        d_data = data
    linestyles = ("-", "--", ":")
    symbols = "x^*"
    for data_2d, linestyle, symbol in zip(d_data, linestyles, symbols):
        for diff in range(n_difference + 1):
            autos = corr_func(data_2d, lags)
            for auto, color in zip(autos.T, "bgrcmykw"):
                ax.plot(
                    auto,
                    color + linestyle + symbol,
                    markerfacecolor=(0, 0, 0, 0),
                    markeredgecolor=color,
                )
            xes = list(range(0, k_range, n_per_day))
            ax.set_xticks(xes, minor=xes)
            if var_names is not None:
                ax.legend(
                    var_names,
                    loc="center left",
                    bbox_to_anchor=(1, 0.5),
                    frameon=False,
                    borderaxespad=0.0,
                )
    ax.grid(True)
    fig.suptitle(title)
    if window_title:
        try:
            fig.canvas.set_window_title(
                "%s (%d)" % (title, fig.canvas.manager.num)
            )
        except AttributeError:
            pass
    return fig, ax


# def cross_cov_nan(data, k):
#    """Biased cross covariance of an array containig nans."""
#    n_vars = data.shape[0]
#    means = my.nanavg(data, axis=1)[:, np.newaxis]
#    cross = np.empty((n_vars, n_vars))
#    for ii in range(n_vars):
#        for jj in range(n_vars):
#            cross[ii, jj] = my.nanavg((data[ii, :-k
def cross_cov(data, k, means=None):
    """Return the cross-covariance matrix for lag k. Variables are assumed to
    be stored in rows, with time extending across the columns."""
    #    if k == 0:
    #        # very funny, jackass
    #        return np.cov(data)
    n_vars, T = data.shape
    k_right = -abs(k) if k else None
    if means is None:
        means = np.nanmean(data, axis=1).reshape((n_vars, 1))
        ddof = 1
    else:
        ddof = 0
    cross = np.empty((n_vars, n_vars))
    for ii in range(n_vars):
        for jj in range(n_vars):
            cross[ii, jj] = np.nansum(
                (data[ii, :k_right] - means[ii]) * (data[jj, k:] - means[jj])
            ) / (T - ddof)
            # cross[ii, jj] = \
            #     np.nanmean((data[ii, :k_right] - means[ii]) *
            #                (data[jj, k:] - means[jj]), ddof=ddof)
    return cross


def cross_corr(data, k):
    """Return the cross-correlation-coefficient matrix for lag k.

    Parameters
    ----------
    data : (K, T) array
        K number of variables. T time steps.
    k : int or iterable
        time-lag or time lags if k is an iterable
    """
    if not isinstance(k, collections.abc.Iterable):
        k = (k,)
    finite_ii = np.isfinite(data)
    stds = [np.nanstd(row[row_ii]) for row, row_ii in zip(data, finite_ii)]
    stds = np.array(stds)[:, np.newaxis]
    stds_dot = stds * stds.T  # dyadic product of row-vector stds
    return np.squeeze([cross_cov(data, lag) / stds_dot for lag in k])


def plot_cross_corr(
    data,
    var_names=None,
    max_lags=10,
    figsize=None,
    fig=None,
    axss=None,
    *args,
    **kwds,
):
    K = data.shape[0]
    if var_names is None:
        var_names = np.arange(K).astype(str)
    lags = np.arange(max_lags)
    # shape: (max_lags, K, K)
    cross_corrs = np.array([cross_corr(data, k) for k in lags])
    if fig is None and axss is None:
        size = {}
        if figsize:
            size["figsize"] = figsize
        fig, axss = plt.subplots(
            K, squeeze=True, sharey=True, sharex=True, **size
        )
    for var_i in range(K):
        lines = []
        for var_j in range(K):
            # want to set the same colors as before when called with a given
            # fig and axss
            colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
            color = colors[var_j % len(colors)]
            lines += axss[var_i].plot(
                lags, cross_corrs[:, var_i, var_j], color=color, *args, **kwds
            )
        axss[var_i].set_title(var_names[var_i])
        axss[var_i].grid(True)
        axss[var_i].set_ylabel("correlation")
    axss[-1].set_xlabel("time lag")
    plt.subplots_adjust(right=0.75, hspace=0.25)
    if fig is not None:
        fig.legend(lines, var_names, loc="center right")
    return fig, axss


def plot_scaling(
    data,
    agg_funcs=(np.mean, np.std, sp.stats.skew),
    var_names=None,
    max_agg_len=365,
    fig=None,
    axs=None,
    *args,
    **kwds,
):
    if var_names is None:
        var_names = np.arange(data.shape[0]).astype(str)
    if fig is None and axs is None:
        fig, axs = plt.subplots(len(var_names), len(agg_funcs), sharex=True)

    # should be of shape:
    # (len(agg_funcs), max_agg_len - 2, len(var_names))
    data_agg = np.array(
        [
            [
                agg_func(
                    [
                        smooth(sub, agg_len, window_function="flat")
                        for sub in data
                    ],
                    # my.sumup(data, agg_len, mean=True),
                    axis=1,
                )
                for agg_len in range(2, max_agg_len)
            ]
            for agg_func in agg_funcs
        ]
    )

    for var_i, var_name in enumerate(var_names):
        for agg_j, agg_func in enumerate(agg_funcs):
            ax = axs[var_i, agg_j]
            ax.semilogx(data_agg[agg_j, :, var_i], *args, **kwds)
            ax.grid(True)
            if var_i == 0:
                ax.set_title(agg_func.__name__)
            if agg_j == 0:
                ax.set_ylabel(var_name)
    return fig, axs


def partial_corr_(x, y, u, rank_cor=True):
    """Is the correlation between x and y because of u?.
    Note that this is only a reasonable measure if x, y and u are normally
    distributed."""
    if rank_cor:
        corr_func = my.kendalls_tau
    else:

        def corr_func(x, y):
            return np.corrcoef(x, y)[0, 1]

    r_xy = corr_func(x, y)
    r_xu = corr_func(x, u)
    r_yu = corr_func(y, u)
    return (r_xy - r_xu * r_yu) / ((1 - r_xu**2) * (1 - r_yu**2)) ** 0.5


def rank_corr_ij(data):
    """Computes the rank correlation coefficient matrix."""
    n_vars = data.shape[0]
    r_ij = np.empty((n_vars, n_vars))
    # fill the lower triangular matrix
    for ii in range(n_vars):
        for jj in range(ii):
            r_ij[ii, jj] = my.kendalls_tau(data[ii], data[jj])
        # do not compute the correlation of a variable with itself
        r_ij[ii, ii] = 1
    # fill the upper triangular matrix, exploiting symmetry
    for ii in range(n_vars):
        if ii + 1 < n_vars:
            for jj in range(ii + 1, n_vars):
                r_ij[ii, jj] = r_ij[jj, ii]
    return r_ij


def cross_rank_corr(data, k):
    """Computes the cross rank correlation for lag k."""
    if k == 0:
        # this should be much faster because it avoids doing double work
        return rank_corr_ij(data)
    n_vars = data.shape[0]
    r_ij = np.empty((n_vars, n_vars))
    for ii in range(n_vars):
        for jj in range(n_vars):
            r_ij[ii, jj] = my.kendalls_tau(data[ii, k:], data[jj, :-k])
    return r_ij


def partial_corr(data, index, diff=False, rank_cor=True, r_ij=None):
    """Returns a matrix with partial (rank) correlation coefficients of all
    variables with respect to the variables in row "index".
    If diff is given as True, the difference to the correlation coefficients
    will be returned.
    If rank_cor is given as False, the ordinary partial correlations are used.
    """
    if r_ij is None:
        if rank_cor:
            r_ij = rank_corr_ij(data)
        else:
            r_ij = cross_corr(data, 0)
    # correlation coefficients between all variables and the one under study
    r_ik = r_ij[index]
    # the same as a column array
    r_kj = r_ik[:, np.newaxis]
    # to iterate is human and to recurse is divine, but to broadcast is to
    # enter Nirwana
    partial = (r_ij - r_ik * r_kj) / ((1 - r_ik**2) * (1 - r_kj**2)) ** 0.5
    if diff:
        partial = r_ij - partial
    return partial


def hurst_coefficient(values):
    """Estimates the Hurst coefficient.

    Parameters
    ----------
    values :    1-dim ndarray
                Data representing a time-series.

    References
    ----------
    http://en.wikipedia.org/wiki/Hurst_coefficient#Estimating_the_exponent
    """
    # find n_partitions, so that the minimum number of observations in a
    # sub-time-series is 8
    N = values.size
    n_partitions = int(np.log2(float(N) / 8))
    Rn_Sn = np.zeros(n_partitions)
    region_sizes = np.zeros(n_partitions)
    for part_i in range(n_partitions):
        # form sub-time-series by reshaping, so we can do all calculation with
        # the help of axis=1
        n_parts = 2**part_i
        cutoff = -(N % n_parts)
        if cutoff == 0:
            cutoff = None
        Xt = values[:cutoff].reshape((n_parts, -1))
        Yt = Xt - Xt.mean()
        Zt = np.cumsum(Yt, axis=1)
        Rn = np.max(Zt, axis=1) - np.min(Zt, axis=1)
        # Sn = np.sqrt(np.mean((Xt - Xt.mean(axis=1)[:, np.newaxis]) ** 2))
        Sn = np.std(Xt, axis=1)
        Rn_Sn[part_i] = np.mean(Rn / Sn)
        region_sizes[part_i] = Xt.shape[1]

    # do a regression on x=log(n), y=log(Rn/Sn)
    return sp.stats.linregress(np.log2(region_sizes), np.log2(Rn_Sn))[0]


def mann_kendall(values, sign_niv=0.025):
    """Test on the significance of a trend.

    Parameters
    ----------
    values :    1-dim ndarray
                Data representing a time-series.
    sign_niv : niveau for the two-sided test

    Returns
    -------
    trend : {1, -1, 0}
        1: significant increasing trend, -1: significant decreasing trend,
        0: no significant trend

    References
    ----------
    http://www.bodensee-hochwasser.info/pdf/langzeitverhalten-bodensee-wasserstaende.pdf
    page 36f
    """
    n = len(values)
    Q = 0
    for i in range(0, n - 1):
        Q = Q + np.sum(np.sign(values[i + 1 :] - values[i]))
    sigma = ((n * (n - 1) * (2 * n + 5)) / 18.0) ** 0.5
    z = Q / sigma
    if z < sp.stats.norm.ppf(sign_niv / 2.0):
        trend = -1  # decreasing trend
    elif z > sp.stats.norm.ppf(1 - sign_niv / 2.0):
        trend = 1  # increasing trend
    else:
        trend = 0  # no significant trend
    return trend


# def test_partial_corr(data, index, alpha=.05, rank_cor=True, r_ij=None):
#    """Returns an array with True/False values whether the Null Hypothesis
#    that x and y are partially uncorrelated can be rejected.
#    See Hartung p.562"""
#    n = data.shape[1]
#    if r_ij is None:
#        r_ij = rank_corr_ij(data)
#    # to only test for partially uncorrelatedness where it makes sense, we
#    # limit ourselves to elements that are rank-correlated at all
#    # see Hartung p.560
#    corr_test_statistic = n * (n - 1) / 2. * r_ij
#    # i have no theoretical values of the test statistic, but the following
#    # transformation puts it into the standard normal world.
#    corr_test_statistic /= (n * (n - 1) * (2. * n + 5) / 18) ** .5
#    corr_mask = (np.abs(corr_test_statistic) >
#                 sp.stats.norm.ppf(1 - .5 * alpha))
#    # approximation of the theoretical K-statistic
#    partial = partial_corr(data, index, rank_cor=rank_cor, r_ij=r_ij)
#    test_statistic = np.abs(partial * (n - 3) ** .5 /
#                            (1 - partial ** 2) ** .5)
# #    return corr_mask & (test_statistic > stats.t.ppf(1 - .5 * alpha, n - 3))
#    return test_statistic > sp.stats.t.ppf(1 - .5 * alpha, n - 3)


if __name__ == "__main__":
    import varwg

    met_vg = varwg.VG(("theta", "ILWR", "Qsw", "rh", "u", "v"))
    met_vg.fit(2)
    simt, sim = met_vg.simulate()
    # met_vg.plot_autocorr()
    met_vg.plot_doy_scatter("theta")



# cython: nonecheck=False
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: language_level=3

from datetime import datetime

import numpy as np
from libc.math cimport sin, cos, pow
# from libc.stdlib cimport malloc
from cython.parallel import prange, parallel
from timezonefinder import TimezoneFinder
import pytz
import bottleneck

import varwg
from varwg.ctimes import datetime2doy

cimport numpy as np

ctypedef np.double_t DTYPE_t

tzf = TimezoneFinder()


def pot_s_rad(date, lat, longt, in_format='%Y-%m-%dT%H:%M', tz_mer=15.0,
            wog=-1):
    """ theoretical maximal potential solar radiation outside of atmosphere

    Parameters
    ----------
    date : numpy.array of time strings (format: in_format) or datetime objects
        or floats (doys)
    lat : float, optional
        latitude of station in decimal degrees, default: Stuttgart Lauchaecker
    longt : float, optional
        longitude of station in decimal degrees, default: Stuttgart Lauchaecker
    in_format : time format string, optional
        format of date if date is string, default '%Y-%m-%dT%H:%M'
    tz_mer : int, optional
        central meridian of time zone, default: 15 (CET)
    wog : {-1, 1}, optional
        west of greenwich, 1 if west, -1 if east, default = -1

    Returns
    -------
    smax : numpy.array of floats
        maximal potential solar radiation in W/m^2

    Notes
    -----
    from campbell technical note 18 [1]_, except declination of sun (d):
    formula of Spencer (1971) [2]_

    WARNING: Campbell Scientific recommends the use of a high quality sun
    screen lotion when exposing your skin to solar radiation for large values
    of sunshine hours!

    References
    ----------
    .. [1] Campbell Scientific (2005) technical note 18: CALCULATING SUNSHINE
       HOURS FROM PYRANOMETER / SOLARIMETER DATA
    .. [2] Spencer JW (1971) Fourier series representation of the position of
       the Sun. Search 2: 172.

    Examples
    --------
    >>> date_str = np.array(["2011-09-28T11:27"])
    >>> pot_s_rad(date_str)
    array([ 855.35624182])
    >>> from varwg import times
    >>> dt = times.str2datetime(date_str, "%Y-%m-%dT%H:%M")
    >>> pot_s_rad(dt)
    array([ 855.35624182])
    >>> pot_s_rad(times.datetime2doy(dt))
    array([ 855.35624182])
    """
    cdef double j1, j2, Et, t0, la, d, sinphi
    cdef int i
    cdef int n = len(date)
    cdef double pi = np.pi
    cdef double[::] smax = np.empty_like(date, dtype=float)
    cdef int s0 = 1373  # Solarkonstante W/m^2
    if tz_mer is None:
        timezone_str = tzf.timezone_at(lat=float(lat),
                                       lng=float(longt))
        timezone = pytz.timezone(timezone_str)
        dt = datetime(2020, 6, 1)
        tz_offset = ((timezone.utcoffset(dt, is_dst=False).total_seconds())
                     / 3600)
        tz_mer = 15 * tz_offset    
    # once upon a time there was a latitude
    cdef double Lc = wog * (tz_mer - longt) / 15.0  # Local correction of time
    lat *= np.pi / 180  # in rad
    try:
        # where we used to raise an exception
        doys = datetime2doy(list(date))  # if date is datetime
    except (TypeError, AttributeError, ValueError, NotImplementedError,
            IndexError):
        try:
            # ...or two
            # if date is string
            doys = np.array(
                datetime2doy(list(varwg.times.str2datetime(date, in_format))))
        except (TypeError, IndexError):
            # those were the doys my friend, i thought they never end
            doys = date  # if date is already in doys
    cdef double[:] hours = (doys - doys.astype(int)) * 24
    cdef long[:] doys_int = doys.astype(int)
    cdef double sinlat = sin(lat)
    cdef double coslat = cos(lat)
    for i in prange(n, nogil=True, schedule="static"):
        j1 = doys_int[i] / 100.
        j2 = (doys_int[i] - 180) / 100.
        if doys_int[i] > 180:
            Et = (-0.05039 - 0.33954 * j2 + 0.04084 * j2 ** 2 + 1.8928
                  * j2 ** 3 - 1.7619 * j2 ** 4 + 0.4224 * j2 ** 5)
        else:
            Et = (-0.04056 - 0.74503 * j1 + 0.08823 * j1 ** 2 + 2.0516
                  * j1 ** 3 - 1.8111 * j1 ** 4 + 0.42832 * j1 ** 5)
        t0 = 12 - Lc - Et
        # we sang and dance forever and a doy (less)
        # la la la la lala lala lala lala
        la = 2 * pi / 365 * (doys_int[i] - 1)
        d = (0.006918 - 0.399912 * cos(la) + 0.070257 * sin(la) -
             0.006758 * cos(2 * la) + 0.000907 * sin(2 * la) -
             0.002697 * cos(3 * la) + 0.00148 * sin(3 * la))
        sinphi = (sin(d) * sinlat +
                  cos(d) * coslat *
                  cos(15 * pi / 180.0 * (hours[i] - t0)))
        if sinphi < 0:
            sinphi = 0
        smax[i] = s0 * sinphi
    return smax


cdef pot_s_rad_c(double[:] doys,
                 float lon,
                 float lat,
                 float tz_mer,
                 float wog):
    cdef double j1, j2, Et, t0, la, d, sinphi, hour
    cdef double sinla, cosla, sin2la, cos2la, cosla2, sin3la, cos3la
    cdef int i, doy_int
    cdef int n = len(doys)
    cdef double pi = np.pi
    cdef double[::] smax = np.empty_like(doys)
    cdef int s0 = 1373  # Solarkonstante W/m^2
    # once upon a time there was a latitude
    # Local correction of time
    cdef double Lc = wog * (tz_mer - lon) / 15.0 
    lat *= np.pi / 180  # in rad
    # cdef double[:] hours = (doys - doys.astype(int)) * 24
    # cdef long[:] doys_int = doys.astype(int)
    cdef double sinlat = sin(lat)
    cdef double coslat = cos(lat)
    cdef double twopi365 = 2 * pi / 365
    # cdef float[:] j2_facts = np.array([-0.05039, -0.33954, 0.04084,
    #                                     1.8928, -1.7619, 0.4224],
    #                                   dtype=np.float32)
    # cdef float[:] j1_facts = np.array([-0.04056, -0.74503, 0.08823,
    #                                     2.0516, -1.8111, 0.42832],
    #                                   dtype=np.float32)
    # for i in prange(n, nogil=True, schedule="static"):
    for i in xrange(n):
        doy_int = <int> doys[i]
        hour = (doys[i] - doy_int) * 24
        j1 = doy_int / 100.
        j2 = (doy_int - 180) / 100.
        # if doy_int > 180:
        #     Et = j2_facts[0] + j2_facts[1] * j2
        #     for i in xrange(2, 6):
        #         j2 *= j2
        #         Et += j2_facts[i] * j2
        # else:
        #     Et = j1_facts[0] + j1_facts[1] * j1
        #     for i in xrange(2, 6):
        #         j1 *= j1
        #         Et += j1_facts[i] * j1
        if doy_int > 180:
            Et = (-0.05039 - 0.33954 * j2 + 0.04084 * j2 ** 2 + 1.8928
                  * j2 ** 3 - 1.7619 * j2 ** 4 + 0.4224 * j2 ** 5)
        else:
            Et = (-0.04056 - 0.74503 * j1 + 0.08823 * j1 ** 2 + 2.0516
                  * j1 ** 3 - 1.8111 * j1 ** 4 + 0.42832 * j1 ** 5)
        t0 = 12 - Lc - Et
        # we sang and dance forever and a doy (less)
        # la la la la lala lala lala lala
        # la = 2 * pi / 365 * (doy_int - 1)
        la = twopi365 * (doy_int - 1)
        sinla = sin(la)
        # vdSin(<int> 1, &la, &sinla)
        cosla = cos(la)
        sin2la = 2 * sinla * cosla
        cosla2 = cosla ** 2
        cos2la = cosla2 - sinla ** 2
        sin3la = sinla * (4 * cosla2 - 1)
        cos3la = 4 * cosla * cosla2 - 3 * cosla
        d = (0.006918 - 0.399912 * cosla + 0.070257 * sinla -
             0.006758 * cos2la + 0.000907 * sin2la -
             0.002697 * cos3la + 0.00148 * sin3la)
        sinphi = (sin(d) * sinlat +
                  cos(d) * coslat *
                  cos(15 * pi / 180.0 * (hour - t0)))
        if sinphi < 0:
            smax[i] = 0
        else:
            smax[i] = s0 * sinphi
    return smax


def sunshine_pot(doys, lat, longt, tz_mer=15.0, wog=-1):
    """Maximum daily sunshine hours based on evaluating pot_s_rad per minute.
    """
    cdef int mins_per_day = 24 * 60
    # doys = doys.astype(float)
    doys = doys[doys == doys.astype(int)]
    doys_minutes = doys.repeat(mins_per_day).reshape(-1, mins_per_day)
    doys_minutes += (np.arange(mins_per_day) /
                     float(mins_per_day))[None, :]
    doys_minutes = doys_minutes.ravel()
    smax = pot_s_rad_c(doys_minutes, longt, lat, tz_mer, wog)
    smax = np.array(smax).reshape(-1, mins_per_day)
    sun_hours = np.sum(smax > 0, axis=1) / 60.
    return sun_hours


cdef inline double mean(np.uint8_t[:] array,
                        unsigned int n) nogil:
    cdef unsigned int i
    cdef double sum = array[0]
    for i in range(1, n):
        sum += array[i]
    return sum / n

# cdef inline brunner_compound_(double[:] Ta_ranks,
#                               double[:] P_ranks,
#                               int n,
#                               double[:] bc):
cdef inline brunner_compound_(np.ndarray[DTYPE_t, ndim=1] Ta_ranks,
                              np.ndarray[DTYPE_t, ndim=1] P_ranks,
                              int n,
                              np.ndarray[DTYPE_t, ndim=1] bc):
    """Rank-based hot-dry index.


    Notes
    -----
    Brunner 2021 uses E-GPD for precipitation and a STI index for
    temperature. This implementation just uses empirical ranks.

    References
    ----------
    Brunner, Manuela I., Eric Gilleland, and Andrew W. Wood.
    “Space–Time Dependence of Compound Hot–Dry Events in the United
    States: Assessment Using a Multi-Site Multi-Variable Weather
    Generator.” Earth System Dynamics 12, no. 2 (May 19, 2021):
    621–34. https://doi.org/10.5194/esd-12-621-2021.
    """
    cdef unsigned int i, j
    cdef double Ta_rank, P_rank
    with nogil:
        for i in range(n):
            Ta_rank = Ta_ranks[i]
            P_rank = P_ranks[i]
            # bc[i] = mean((Ta_ranks <= Ta_rank) & (P_ranks <= P_rank),
            #              n)
            for j in range(n):
                # if (Ta_ranks[j] <= Ta_ranks[i]) and (P_ranks[j] <= P_ranks[i]):
                if (Ta_ranks[j] <= Ta_rank) and (P_ranks[j] <= P_rank):
                    bc[i] += 1
            bc[i] /= n

def brunner_compound(Ta, P, progress=False):
    if progress:
        from tqdm import tqdm as progress
    else:
        def progress(x, *args, **kwds):
            return x
    cdef unsigned int row_i
    Ta_ranks = bottleneck.rankdata(Ta, axis=-1)
    P_ranks = bottleneck.rankdata(np.negative(P), axis=-1)
    shape_before = Ta_ranks.shape
    cdef int T = shape_before[len(shape_before) - 1]
    Ta_ranks = Ta_ranks.reshape(-1, T)
    P_ranks = P_ranks.reshape(-1, T)
    cdef int n_rows = len(Ta_ranks)
    bc = np.zeros_like(Ta_ranks)
    for row_i in progress(range(n_rows), total=n_rows):
        brunner_compound_(Ta_ranks[row_i], P_ranks[row_i], T, bc[row_i])
    bc = bc.reshape(shape_before)
    return bc




import copy
import datetime
import functools
import os
import pickle
import shutil
import warnings
from collections import defaultdict, namedtuple
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import varwg
from varwg import helpers as my
from varwg.core import base, plotting
from varwg.meteo import avrwind, meteox2y

# from varwg.meteo.meteox2y_cy import pot_s_rad
from varwg.meteo.meteox2y import pot_s_rad, sunshine_riseset

# from varwg.time_series_analysis import (
#     conditional_sim as csim,
# )
from varwg.time_series_analysis import (
    cresample,
    distributions,
    models,
)
from varwg.time_series_analysis import (
    resample as resampler,
)
from varwg.time_series_analysis import seasonal_distributions as sd
from varwg.time_series_analysis import (
    time_series as ts,
)

read_met = base.read_met

try:
    from varwg import config as conf
except ImportError:
    from varwg import config_template as conf

    conf_filepath = conf.__file__
    if conf_filepath.endswith(".pyc"):
        conf_filepath = conf_filepath[:-1]


def delete_cache():
    shutil.rmtree(conf.cache_dir)


def dump_data(
    times_,
    data,
    var_names,
    p,
    q,
    extra="",
    random_state=None,
    out_dir=None,
    conversions=None,
):
    """Dumps the given datetimes and data in an ascii file and random state in
    a pickle file."""
    if out_dir is None:
        out_dir = conf.out_dir
    if p is None:
        p = 0
    if conversions:
        data = np.copy(data)
        for conversion in list(conversions):
            times_, data, var_names = conversion(times_, data, var_names)
    outfilename = "VARMA_p%d_q%d_sim%s.dat" % (p, q, extra)
    outfilepath = os.path.join(out_dir, outfilename)
    with open(outfilepath, "w") as txt_file:
        txt_file.write("time\t" + "\t".join(var_names) + os.linesep)
        for time, values in zip(times_, data.T):
            txt_file.write(
                "%s\t%s%s"
                % (
                    time.isoformat(),
                    "\t".join(
                        conf.out_format[var_name] % val
                        for var_name, val in zip(var_names, values)
                    ),
                    os.linesep,
                )
            )

    if random_state is not None:
        with open(outfilepath + ".random_state", "wb") as pi_file:
            pickle.dump(random_state, pi_file)

    return outfilepath


def metfile2df(met_file):
    try:
        pd_kwds = dict(
            skiprows=5,
            index_col=0,
            parse_dates=True,
            sep="\t",
            date_parser=varwg.times.cwr2datetime,
        )
        met_df = pd.read_csv(met_file, **pd_kwds).astype(float)
    except TypeError:

        def date_hour2datetime(date_str, hour_str):
            return varwg.times.str2datetime(
                "%s %s" % (date_str, hour_str), "%d. %m %y %H"
            )

        pd_kwds = dict(
            index_col=0, parse_dates=[[0, 1]], date_parser=date_hour2datetime
        )
        met_df = pd.read_csv(met_file, **pd_kwds).astype(float)
    return met_df


def outfile2df(outfile):
    pd_kwds = dict(index_col=0, parse_dates=[0], sep="\t")
    return pd.read_csv(outfile, **pd_kwds).astype(float)


def metfile2hdf5(met_file, h5_filename=None, key="met"):
    if h5_filename is None:
        h5_filename = os.path.splitext(met_file)[0] + ".h5"
    met_df = metfile2df(met_file)
    met_df.to_hdf(h5_filename, key)


def seasonal_back(
    dist_sol,
    norm_data,
    var_names,
    doys=None,
    solution_template="%s",
    pass_doys=True,
    var_names_back=None,
    mean_shifts=None,
):
    """Transform variables from normal to real-world marginals.

    Parameters
    ----------
    dist_sol : mapping like dict or shelve object
        Distribution fitting parameters.
    norm_data : (K, T) array
        Normally distributed data.
    var_names : (K,) sequence of str
        Variable names used as dist_sol[solution_template % var_name].
    doys : None or (T,) float array, optional
        Only needed for different doys than the calibration dataset.
    solution_template : str, optional
        Used in the form dist_sol[solution_template % var_name]

    Returns
    -------
    data : (K, T) array
        Re-transformed data
    """
    data = np.empty_like(norm_data)
    if mean_shifts is None:
        mean_shifts = defaultdict(lambda: None)
    if var_names_back is None:
        var_names_back = var_names
    for var_name in var_names_back:
        var_i = var_names.index(var_name)
        var = norm_data[var_i]
        # for var_i, (var_name, var) in enumerate(zip(var_names, norm_data)):
        distribution, solution = dist_sol[solution_template % var_name]
        if hasattr(distribution, "dist") and isinstance(
            distribution.dist, distributions.Normal
        ):
            # in this case, we do not need a qenuine qq-transform, the normal
            # inverse Z transform suffices, AND is immune to producing nans
            # for "extreme" values in the standard-normal world!
            if doys is not None:
                _T = (2 * np.pi / 365 * doys)[np.newaxis, :]
            else:
                _T = None
            mus, sigmas = distribution.trig2pars(solution, _T)
            data[var_i] = var * sigmas + mus
        else:
            quantiles = distributions.norm.cdf(var)
            doys_ = doys if pass_doys else None
            data[var_i] = distribution.ppf(
                solution,
                quantiles,
                doys=doys_,
                mean_shift=(mean_shifts.get(var_name, None)),
            )
    return data


def sim_byvariable(values, times, p=2, aggregation="%w"):
    aggr_values, aggr_times = ts.time_part_average(
        values, times, aggregation, expand=False, return_times=True
    )
    dist = sd.SeasonalDist(distributions.norm, aggr_values, aggr_times)
    solution = dist.fit(T0=50)
    trans = distributions.norm.ppf(dist.cdf(solution), 0, 1)
    B, sigma_u = models.VAR_LS(trans, p=p)
    T = len(aggr_values)
    sim = models.VAR_LS_sim(B, sigma_u, T - p)
    retrans = dist.ppf(solution, distributions.norm.cdf(sim, 0, 1))
    retrans = np.repeat(retrans, varwg.times.time_part_average.repeats)
    return retrans


def interannual_variability(T, mean_arrival=30, disturbance_std=0.1):
    """Generate a time series used as a disturbance within the std-norm world.
    Disturbances are simulated as gaussian changes occuring after t timesteps,
    where t is drawn from an exponential distribution.
    """

    def arrival():
        return distributions.expon.ppf(
            varwg.rng.random(),
            lambd=1.0 / mean_arrival,
            # loc=1 weil wir taegliche Werte
            # berechnen und daher 1 Tag die
            # kuerzest moegliche Dauer ist
            x0=1,
        )

    def disturbance():
        return abs(
            distributions.norm.ppf(
                varwg.rng.random(),
                sigma=float(disturbance_std),
            )
        )

    m_t = np.zeros(T)
    breaks_ii = [t] = [int(round(arrival()))]
    while t < T:
        breaks_ii += [breaks_ii[-1] + int(round(arrival()))]
        t = breaks_ii[-1]
    sign = 1
    for break_, next_break in zip(breaks_ii[:-1], breaks_ii[1:]):
        period_length = min(int(round(next_break - break_)), len(m_t) - break_)
        window = np.bartlett(period_length + 1)[:-1]
        if window.mean() <= 0:
            window = np.ones(period_length) / period_length
        else:
            window /= window.mean()
        m_t[break_:next_break] = sign * disturbance() * window
        sign *= -1
    return m_t


def sw_diurnal(date, daily_sw_data, del_t=3600):
    """add standard daily cycle to daily mean sw data
    Parameters
    ----------
    date: array
        date in unix timestamps
    daily_sw_data: array
    del_t: int
        timestep of output data in seconds, default = 3600s

    Returns
    -------
    date_: array
        output date in unix timestamps
    data_: array
        sw data in del_t-steps"""
    nn = 86400.0 / del_t
    date_ = np.arange(date[0], date[-1] + 86400, del_t)
    swmax = pot_s_rad(
        varwg.times.unix2str(date_, "%Y-%m-%dT%H:%M"),
        lat=conf.latitude,
        longt=conf.longitude,
    )
    swmax_daily = np.average(swmax.reshape(-1, nn), axis=1).repeat(nn)
    fkt = swmax / swmax_daily
    data_ = daily_sw_data.repeat(nn) * fkt
    return date_, data_


class VarWG(plotting.Plotting):
    """A Vector-Autoregressive weather generator.

    >>> my_vg = VG(("theta", "Qsw", "ILWR", "rh", "u", "v"))
    >>> my_vg.fit()
    >>> times_out, sim_data = my_vg.simulate()
    """

    def __init__(
        self,
        var_names,
        met_file=None,
        sum_interval=24,
        max_nans=12,
        plot=False,
        separator="\t",
        refit=None,
        detrend_vars=None,
        verbose=True,
        data_dir=None,
        cache_dir=None,
        dump_data=True,
        non_rain=None,
        rain_method="regression",
        conf_update=None,
        station_name=None,
        infill=False,
        fit_kwds=None,
        **met_kwds,
    ):
        """A vector autoregressive moving average (VARMA) weather generator.
        Initializing a VG-object does the following:
            - read input data from a met-file
            - aggregate it according to sum_interval
            - convert the data using seasonal distributions

        To fit the VARMA process call fit(). Afterwards
        simulated time series can be obtained with simulate().

        Parameters
        ----------
        var_names : sequence containing strings
        met_file : string, dict, pandas DataFrame or None, optional
            If None, value will be read from config.py.
            Data can also be supplied directly with a dict, mapping variable
            names to ndarrays containing the data. The dictionary has to
            contain an item with the key "datetimes" which has to contain an
            ndarray of datetime objects.
            If data is supplied as a pandas DataFrame, time information will be
            taken from its index.
        sum_inverval : int or sequence of ints, optional
            Number of timesteps to aggregate. If a sequence is given, its
            elements are used variable - specific. In this case the sequence
            has to have the same length as var_names.
            Default 24 corresponds to the case of hourly input data and daily
            simulation
        plot : boolean, optional
            Plot all kinds of information. Warning: a lot of windows will
            appear.
        separator : str or None, optional
            String that separates columns in the met_file. It is fed to the
            split method of str, so None corresponds to any whitespace.
        refit : None, 'all' or sequence of strings from var_names, optional
            Instead of using cached solutions for the seasonal distribution
            fit, run the fit again and cache it.
        detrend_vars : sequence containing strings or None, optional
            Remove a linear trend from the variables.
        verbose : boolean, optional
            Be more verbose if True.
        data_dir : str, optional
            Overwrites the data_dir variable from the config-file.
        cache_dir : str, optional
            Overwrites the cache_dir variable from the config-file.
        dump_data : boolean, optional
            Write simulation and disaggregation results (filename can be
            infered by the outfilepath attribute).
        non_rain : None or sequence of str, optional
            Variables used for rain gap filling.
        rain_method : 'distance' or 'regression', optional
            Which method to use for rain gap filling. 'distance' uses
            euclidean distance to wet states. 'regression' uses a
            regression to extrapolate from wet do dry amounts.
        conf_update : dict or None, optional
            Replaces variables in the config file.
        station_name : str or None, optional
            Name of the station. Usefull in combination with
            weathercop multisite generation.
        infill : boolean, optional
            VAR-based infilling of missing values.
        """
        # external_var_names=None,
        # "fix" unicode problems
        var_names = [str(var_name) for var_name in var_names]
        self.station_name = station_name
        if conf_update is not None:
            self._conf_update(conf_update)
        super(VarWG, self).__init__(
            var_names,
            met_file=met_file,
            sum_interval=sum_interval,
            plot=plot,
            separator=separator,
            refit=refit,
            detrend_vars=detrend_vars,
            verbose=verbose,
            data_dir=data_dir,
            cache_dir=cache_dir,
            seasonal_fitting=True,
            dump_data=dump_data,
            non_rain=non_rain,
            rain_method=rain_method,
            max_nans=max_nans,
            infill=infill,
            fit_kwds=fit_kwds,
            **met_kwds,
        )
        # simulated residuals
        self.ut = None

    def _conf_update(self, conf_update):
        for name, value in conf_update.items():
            setattr(conf, name, value)
            setattr(base.conf, name, value)
            setattr(plotting.conf, name, value)
            if hasattr(self, name):
                setattr(self, name, value)

    def __getattribute__(self, name):
        # Plotting overwrites __getattribute__ to do a seasonal fitting
        # only if it is (implicitly) requested. That behaviour is only usefull
        # when creating a Plotting instance
        return base.Base.__getattribute__(self, name)

    def __str__(self):
        return f"VG({self.station_name}, {self.rain_method})"

    def fit(
        self,
        p=None,
        q=0,
        p_max=10,
        seasonal=True,
        ex=None,
        ex_kwds=None,
        extro=False,
        doy_width=60,
        fft_order=2,
        *args,
        **kwds,
    ):
        """p is the order of autoregressive process.
        q the order of the moving average process
        If p is not given, it is selected using the Schwartz Information
        Criterion."""
        self.p, self.q, self.seasonal = p, q, seasonal
        self.ex_in = ex
        if q == 0:
            if p is None:
                if self.verbose:
                    print("Perfoming VAR order selection: ", end=" ")
                try:
                    if ex is None:
                        p = models.VAR_order_selection(self.data_trans, p_max)
                    else:
                        est_kwds = {} if ex_kwds is None else ex_kwds
                        est_kwds["ex"] = ex
                        o_select = models.VAR_order_selection
                        p = o_select(
                            self.data_trans,
                            p_max,
                            estimator=models.VAREX_LS,
                            est_kwds=est_kwds,
                        )
                    self.p = p
                except ValueError:
                    warnings.warn(
                        "Could not fit the VAR process. Trying to "
                        "remove extreme values."
                    )
                    mi, ma = np.min(self.data_trans), np.max(self.data_trans)
                    if ma > -mi:
                        self.data_trans.ravel()[
                            np.argmax(self.data_trans)
                        ] = np.nan
                    else:
                        self.data_trans.ravel()[
                            np.argmin(self.data_trans)
                        ] = np.nan
                    if not self.infill:
                        self.data_trans = my.interp_nonfin(self.data_trans)
                    self.fit(p, q, p_max, seasonal, ex, ex_kwds)
                if self.verbose:
                    print("p=%d seems parsimonious to me" % self.p)
                    if p == 0:
                        print(
                            "Wow, you should not do any autoregressive "
                            "modeling."
                        )
            # fit a VAR-model
            if self.seasonal:
                if self.verbose:
                    print("Fitting the seasonal VAR model.")
                self.svar_doy_width = doy_width
                self.svar_fft_order = fft_order
                self.Bs, self.sigma_us = models.SVAR_LS(
                    self.data_trans,
                    self.data_doys,
                    self.p,
                    var_names=self.var_names,
                    verbose=self.verbose,
                    doy_width=doy_width,
                    fft_order=fft_order,
                    *args,
                    **kwds,
                )
                self.residuals = models.SVAR_residuals(
                    self.data_trans, self.data_doys, self.Bs, self.p
                )
            else:
                if self.verbose:
                    print("Fitting the VAR model.")
                if extro:

                    def transforms(x):
                        sim_sea = seasonal_back(
                            self.dist_sol, x, self.var_names, self.data_doys
                        )
                        return sim_sea / self.sum_interval

                    def backtransforms(x):
                        return self._fit_seasonal(values=x, filter_nans=False)[
                            0
                        ]

                    B, self.sigma_u = models.VAR_LS_extro(
                        self.data_trans,
                        (self.data_raw / self.sum_interval),
                        transforms,
                        backtransforms,
                        p=self.p,
                        *args,
                        **kwds,
                    )
                    self.residuals = models.VAR_LS_extro.residuals
                elif ex is None:
                    B, self.sigma_u = models.VAR_LS(
                        self.data_trans, self.p, *args, **kwds
                    )
                    self.residuals = models.VAR_residuals(
                        self.data_trans, B, self.p
                    )
                else:
                    B, self.sigma_u = models.VAREX_LS(
                        self.data_trans, p, ex, *args, **kwds
                    )
                # we call it AM to have one place for VARMA parameters
                self.AM = B
            self.p = p

        else:
            if (p is None) and (q is None):
                if self.verbose:
                    print("Perfoming VARMA order selection: ", end=" ")
                self.p, self.q = p, q = models.VARMA_order_selection(
                    self.data_trans, p_max, plot_table=self.plot, *args, **kwds
                )
                if self.verbose:
                    print("p=%d and q=%d seems parsimonious to me" % (p, q))
            else:
                self.p, self.q = p, q
            # fit a VARMA-model
            if self.verbose:
                print("Fitting the VARMA model")
            # the VARMA implementation is for mean-adjusted data. the data
            # should have zero mean, asuming the fit for the qq-transformation
            # was perfect.
            # ... and nothing is perfect. so we subtract the mean here
            self.means = self.data_trans.mean(axis=1)[:, np.newaxis]
            self.data_trans -= self.means

            AM, sigma_u, residuals = models.VARMA_LS_prelim(
                self.data_trans, p, q, *args, **kwds
            )

            self.data_trans += self.means
            self.AM, self.sigma_u, self.residuals = AM, sigma_u, residuals

    def _gen_online_transform(self, dist_sol, solution_template="%s"):
        """Generates the callables for time_series_analysis.models.VAR_LS_sim
        that enforce the factor-change in precipitation."""
        r_index = self.var_names.index("R")
        r_dist, solution = dist_sol[solution_template % "R"]

        def transform(value_vector, t):
            # back-transform precipitation, apply the real-world
            # factor and transform back to standard-normal
            r_stdn = value_vector[r_index]
            doy = self.sim_doys[t]
            quantile = distributions.norm.cdf(r_stdn, 0, 1)
            r_real = r_dist.ppf(solution, quantile, doys=doy)
            value_vector[r_index] = r_real * self.r_fact
            return value_vector

        return transform

    def simulate(
        self,
        T=None,
        *,
        mean_arrival=None,
        disturbance_std=None,
        theta_incr=None,
        r_fact=None,
        theta_grad=None,
        fixed_variables=None,
        primary_var="theta",
        scale_prim=True,
        climate_signal=None,
        start_str=None,
        stop_str=None,
        random_state=None,
        ex=None,
        ex_kwds=None,
        seed_before_sim=None,
        # loc_shift=False,
        resample=False,
        res_kwds=None,
        asy=False,
        residuals=None,
        sim_func=None,
        sim_func_kwds=None,
        phase_randomize=False,
        phase_randomize_vary_mean=True,
        return_rphases=False,
        rphases=None,
        p_kwds=None,
        taboo_period_min=None,
        taboo_period_max=None,
        conversions=None,
    ):
        """Simulate based on data from __init__ and the fit from calling
        self.fit().
        To choose a specific order of AR and MA call fit(p, q) before calling
        simulate. Otherwise fit will be called without parameters, so p and q
        will be guessed by applying an information criterion.

        Parameters
        ----------
        T : None or int, optional
            Number of timesteps to simulate. T=None produces as many simulation
            timesteps as there were in the data used for fitting the VARMA
            process.
        mean_arrival : float, optional
            average length of warm/cold periods [d]
        disturbance_std : float, optional
            standard deviation of deviation of warm/cold periods from mean
        theta_incr : float, optional
            increase in average of temperature [deg C]
        r_fact : float, optional
            factor to increase precipitation
        theta_grad : float, optional
            gradient increase in average of temperature [deg C / (T)]
        fixed_variables : dictionary mapping var_names to None or (T,) ndarrays
            Keeps the provided time - series fixed. Mappings to None indicate
            that the input data should be used as a fixed variable.
            Can be used to simulate hierarchically.
        primary_var : str, one of var_names from __init__ or sequence of those,
                optional
            All disturbances (mean_arrival, distrubance_std, theta_incr,
            theta_grad and climate_signal) correspond to changes in this
            variable.
        scale_prim : boolean, optional
            Allow primary variables to influence each other.
            Only effective when using multiple primary variables!
        climate_signal : (T,) ndarray
            A time series of the 'primary_var'. Differences between
            'climate_signal' and the seasonal mean are used to perturb the
            VAR process.
        start_str : str of format '%m-%d %H', optional
            String representation of the start date of the simulation.
        stop_str : str of format '%m-%d %H', optional
            String representation of the end date of the simulation.
        random_state : str, tuple as returned by numpy.random.get_state() or
                       None, optional
            Sets the random state before simulating.
            *str* is interpreted as a pickle file containing the random state.
            *tuple* is interpreted as the random state itself.
        ex : (T,) ndarray or function
            External variable. If given as a function, ex_t will be generated
            by calling ex(Y[:t], **ex_kwds), with Y being the simulated values.
        ex_kwds : None or dict
            Keyword arguments to be passed to ex.
        seed_before_sim : None or int, optional
            Seed the Random Generator before simulating. The integer
            passed will be used as seed (also if it is a 0).
            (_scenario_parameters also draws random numbers).
        loc_shift : bool, optional
            Change on the primary variables will be removed after simulation
            and added to the backtransformed variables.
        res_kwds : dict or None, optional
            Keywords for the resampler, if not None, use resampling.
            See `time_series_analysis.resample.resample` for allowed
            parameters.
        residuals : (K, T+) ndarray
            Residuals with which to run the time series model. If more
            than T columns are supplied the additional time steps (at
            the beginning) will be used for spin-up.
        sim_func : None or callable, optional
            Callback to replace the time series model.
        sim_func_kwds : None or dict
            Extra keyword arguments for sim_func
        conversions : iterable of callables

        Returns
        -------
        sim_times : (T,) ndarray of datetime objects
        sim_sea : (K, T) float ndarray

        """
        if sim_func_kwds is None:
            sim_func_kwds = dict()
        if isinstance(random_state, str):
            with open(random_state, "rb") as pi_file:
                random_state = pickle.load(pi_file)
        if random_state is not None:
            varwg.rng.bit_generator.state = random_state
        # we record the random state to later dump it
        pre_sim_random_state = varwg.rng.bit_generator.state

        if mean_arrival is None and disturbance_std is not None:
            raise RuntimeError(
                "Must specify mean_arrival if disturbance_std is given."
            )
        if disturbance_std is None and mean_arrival is not None:
            raise RuntimeError(
                "Must specify disturbance_std if mean_arrival is given."
            )
        (
            self.disturbance_std,
            self.theta_incr,
            self.r_fact,
            self.theta_grad,
            self.mean_arrival,
        ) = [
            None if x is None else np.atleast_1d(x).astype(float)
            for x in (
                disturbance_std,
                theta_incr,
                r_fact,
                theta_grad,
                mean_arrival,
            )
        ]
        self.fixed_variables = fixed_variables
        if primary_var is None:
            # we got to have one
            self.primary_var = primary_var = self.var_names[0]
        if isinstance(primary_var, str):
            self.primary_var = (primary_var,)
            self.climate_signal = np.atleast_1d(climate_signal)
            # self.climate_signal = np.atleast_2d(climate_signal)
        else:
            self.primary_var = primary_var
            self.climate_signal = (
                np.array([None] * len(primary_var))
                if climate_signal is None
                else climate_signal
            )
        # if self.climate_signal.dtype == object:
        #     # we have to back up, this has not worked
        #     self.climate_signal = tuple(self.climate_signal[0])
        for var_name in self.primary_var:
            if var_name not in self.var_names:
                raise ValueError(f"{primary_var=} not in {self.var_names=}")
        if return_rphases and not phase_randomize:
            warnings.warn(
                "Phases were requested, yet phase randomization is not enabled.\n"
                "Setting phase_randomize=True now"
            )
            phase_randomize = True
        if rphases and not phase_randomize:
            warnings.warn(
                "Phases were supplied, yet phase randomization is not enabled.\n"
                "Setting phase_randomize=True now"
            )
            phase_randomize = True
        self.phase_randomize = phase_randomize
        self.phase_randomize_vary_mean = phase_randomize_vary_mean

        if T is None:
            if climate_signal is not None:
                # if we have multiple primary variables, elements of climate
                # signal can be None. we also have to be careful, because the
                # number of primary variables could be interpreted as T
                # climate_signal = np.atleast_2d(climate_signal)
                _signal = np.array(
                    [_ for _ in self.climate_signal if _ is not None]
                )
                self.T_sim = np.atleast_2d(_signal).shape[1]
            else:
                self.T_sim = self.T_summed
        else:
            self.T_sim = T

        # we depend on order selection if self.fit was not called
        if self.p is None and sim_func is None:
            self.fit(**self.fit_kwds)

        # if this is True, we still call self._gen_sim_times because
        # of expected side-effects
        self.sim_times_is_times = start_str is None and stop_str is None
        self.sim_times = self._gen_sim_times(
            start_str=start_str, stop_str=stop_str
        )
        # self.T_sim has to be reset when self._gen_sim_times is
        # called, but not inside self.disaggregate
        self.T_sim = len(self.sim_times)

        if self.phase_randomize:
            if self.theta_incr is None:
                self.theta_incr = np.zeros(len(self.primary_var))
            if residuals is None:
                residuals = self.residuals

        # converts the given parameters so they are understood in the
        # std-normal world
        # to store effect of scenario parameters by primary variable
        self._m_single, self._m_t_single, self._m_trend_single = [], [], []
        m, m_t, m_trend = sc_pars = self._scenario_parameters(
            theta_incr,
            theta_grad,
            disturbance_std,
            climate_signal,
            phase_randomize,
            scale_prim,
            primary_var_sim=sim_func_kwds.get("primary_var_sim", None),
        )
        # store to be able to play outside with it later
        self.m, self.m_t, self.m_trend = m, m_t, m_trend

        fixed_data = self._prepare_fixed_data()

        # relative changes have to be dealt with differently: the
        # factor is dependent on the distribution and therefore can
        # only be done when we now the value of the to-be-transformed
        # variable.
        if r_fact:
            transform = self._gen_online_transform(self.dist_sol)
        else:
            transform = None

        if self.verbose:
            print("Simulating a time-series.")
        if type(seed_before_sim) is int:
            varwg.reseed(seed_before_sim)
        var_names_back = None
        if sim_func is not None:
            if sim_func_kwds is None:
                sim_func_kwds = {}
            sim = sim_func(self, sc_pars, **sim_func_kwds)
            # this is weathercop-specific
            if (stop_at := sim_func_kwds.get("stop_at", None)) is not None:
                vine = sim_func_kwds["wcop"].vine
                var_names_back = vine.varnames[: stop_at + 1]
            if return_rphases := sim_func_kwds.get("return_rphases", False):
                sim, ranks_sim, rphases = sim
            else:
                sim, ranks_sim = sim
        elif resample or res_kwds is not None:
            # in contrast to the parametric models, we do not
            # transform anything
            # here we combine any change in theta_incr
            m_resampler = (m + m_t + m_trend[:, None])[self.primary_var_ii]
            # m_resampler = m[:, 0]
            try:
                res_kwds = my.ADict(res_kwds)
            except TypeError:
                res_kwds = my.ADict()
            cy = "cy" in res_kwds and res_kwds["cy"]
            resample_raw = res_kwds.pop("resample_raw", False)
            if cy:
                sim, self.res_indices, self.candidates = cresample.resample(
                    (
                        self.data_raw / self.sum_interval
                        if resample_raw
                        else self.data_trans
                    ),
                    self.times,
                    self.p,
                    n_sim_steps=self.T_sim,
                    theta_incr=theta_incr if resample_raw else m_resampler,
                    theta_i=self.primary_var_ii,
                    # cache_dir=conf.cache_dir,
                    cache_dir=self.cache_dir,
                    verbose=self.verbose,
                    return_candidates=True,
                    z_transform=resample_raw,
                    **(res_kwds - "cy"),
                )
            else:
                sim, self.res_indices, self.candidates = resampler.resample(
                    data=(
                        self.data_raw / self.sum_interval
                        if resample_raw
                        else self.data_trans
                    ),
                    dtimes=self.times,
                    p=self.p,
                    n_sim_steps=self.T_sim,
                    theta_incr=theta_incr if resample_raw else m_resampler,
                    bias=None,
                    theta_i=self.primary_var_ii,
                    # cache_dir=conf.cache_dir,
                    cache_dir=self.cache_dir,
                    verbose=self.verbose,
                    # verbose=True,
                    return_candidates=True,
                    z_transform=resample_raw,
                    **res_kwds,
                )
        elif self.q in (0, None):
            # simulate VAR-time-series
            if self.seasonal:
                # scale u to not get too much variance
                # n_unique_doys = len(np.unique(self.data_doys))
                # uu = self.svar_doy_width / n_unique_doys * residuals
                uu = residuals
                sim = models.SVAR_LS_sim(
                    self.Bs,
                    self.sigma_us,
                    self.sim_doys,
                    m,
                    ia=m_t,
                    m_trend=m_trend,
                    fixed_data=fixed_data,
                    u=uu,
                    phase_randomize=phase_randomize,
                    return_rphases=return_rphases,
                    rphases=rphases,
                    p_kwds=p_kwds,
                    taboo_period_min=taboo_period_min,
                    taboo_period_max=taboo_period_max,
                )
                if return_rphases:
                    sim, rphases = sim
                self.ut = models.SVAR_LS_sim.ut
            else:
                if asy:
                    if isinstance(asy, str):
                        asy = [self.var_names.index(asy)]
                    elif asy:
                        asy = self.var_names
                    sim = models.VAR_LS_sim_asy(
                        self.AM,
                        self.sigma_u,
                        self.T_sim,
                        self.data_trans,
                        self.p,
                        skewed_i=asy,
                        verbose=self.verbose,
                        var_names=self.var_names,
                        u=residuals,
                    )
                    self.ut = models.VAR_LS_sim_asy.ut
                elif ex is None:
                    sim = models.VAR_LS_sim(
                        self.AM,
                        self.sigma_u,
                        self.T_sim,
                        m,
                        ia=m_t,
                        m_trend=m_trend,
                        fixed_data=fixed_data,
                        transform=transform,
                        u=residuals,
                        phase_randomize=phase_randomize,
                        rphases=rphases,
                        return_rphases=return_rphases,
                        p_kwds=p_kwds,
                        taboo_period_min=taboo_period_min,
                        taboo_period_max=taboo_period_max,
                    )
                    if return_rphases:
                        sim, rphases = sim
                else:
                    sim, self.ex_out = models.VAREX_LS_sim(
                        self.AM,
                        self.sigma_u,
                        self.T_sim,
                        ex,
                        m,
                        ia=m_t,
                        m_trend=m_trend,
                        ex_kwds=ex_kwds,
                        u=residuals,
                    )
        else:
            sim = models.VARMA_LS_sim(
                self.AM,
                self.p,
                self.q,
                self.sigma_u,
                # process means that are used as
                # starting values there is some
                # confusion here...
                self.means,
                self.T_sim,
                m,
                ia=m_t,
                m_trend=m_trend,
                fixed_data=fixed_data,
                u=residuals,
            )

        if self.plot and self.sigma_u is not None:
            ts.matr_img(
                self.sigma_u,
                "Noise Covariance matrix",
                self.var_names,
                self.var_names,
            )

        # # location shifting
        # # DO NOT USE
        # if loc_shift:
        #     __import__("pdb").set_trace()
        #     sim = self._location_shift_normal(sim)

        # transform back
        if resample or res_kwds is not None:

            # use the indices to map back, we do not want any new
            # values to occur here!

            sim_sea = np.array(
                [
                    self.data_raw[
                        self.var_names.index(var_name), self.res_indices
                    ]
                    for var_name in self.var_names
                ]
            )
            sim_sea /= self.sum_interval

            # data_raw = self.data_raw / self.sum_interval
            # sim_sea = (
            #     sim * data_raw.std(axis=1)[:, None]
            #     + data_raw.mean(axis=1)[:, None]
            # )

            # sim_sea = seasonal_back(
            #     self.dist_sol, sim, self.var_names, doys=self.sim_doys
            # )
            # sim_sea /= self.sum_interval

            # if theta_incr > 0:
            #     plt.scatter(self.times, self.data_raw[0] / 24)
            #     plt.scatter(self.times[self.res_indices],
            #                 self.data_raw[0, self.res_indices] / 24)
            #     plt.show()
        else:
            # sim_sea = seasonal_back(self.dist_sol, sim, self.var_names,
            #                         doys=self.sim_doys)
            # sim_sea /= self.sum_interval
            # if "R" in self.var_names:
            #     r_index = self.var_names.index("R")
            #     sim_sea[r_index] *= self.sum_interval[r_index]

            if self.theta_incr is not None:
                mean_shifts = dict(
                    theta=self.theta_incr
                    * self.sum_interval.ravel()[self.primary_var_ii]
                )
            else:
                mean_shifts = None
            sim_sea = seasonal_back(
                self.dist_sol,
                sim,
                self.var_names,
                doys=self.sim_doys,
                var_names_back=var_names_back,
                mean_shifts=mean_shifts,
                # pass_doys=(not self.sim_times_is_times)
            )
            sim_sea /= self.sum_interval

        # spicyness can lead to infs
        sim_sea[~np.isfinite(sim_sea)] = np.nan
        sim_sea = my.interp_nonfin(sim_sea, max_interp=3)

        # if loc_shift:
        #     sim_sea = self._location_shift_back(sim_sea)

        if conversions:
            for conversion in list(conversions):
                self.sim_times, sim_sea, self.var_names_conv = conversion(
                    self.sim_times, np.copy(sim_sea), self.var_names
                )

        # lets store this so we can play with it from the outside
        self.sim, self.sim_sea = sim, sim_sea

        if self.verbose:
            self.print_means()

        if self.dump:
            extra = "" if self.station_name is None else self.station_name
            self.outfilepath = dump_data(
                self.sim_times,
                self.sim_sea,
                self.var_names,
                self.p,
                0 if self.q is None else self.q,
                extra=extra,
                random_state=pre_sim_random_state,
                out_dir=self.data_dir,
                conversions=conversions if conversions else conf.conversions,
            )

        if sim_func is not None:
            if return_rphases:
                return self.sim_times, sim_sea, ranks_sim, rphases
            return self.sim_times, sim_sea, ranks_sim

        if return_rphases:
            return self.sim_times, sim_sea, rphases
        return self.sim_times, sim_sea

    def print_means(self):
        obs = self.to_df("daily input").mean(axis=0)
        obs.name = "obs"
        sim = self.to_df("daily output", with_conversions=False).mean(axis=0)
        sim.name = "sim"
        diff = pd.DataFrame(
            sim.values - obs.values, index=obs.index, columns=["diff"]
        )
        diff_perc = pd.DataFrame(
            100 * (sim.values - obs.values) / obs.values,
            index=obs.index,
            columns=["diff [%]"],
        )
        print(pd.concat([obs, sim, diff, diff_perc], axis=1).round(3))

    # def disaggregate_rm(self, refit=None):
    #     """Random Mixing variant of disaggregation."""
    #     if self.sim is None:
    #         raise RuntimeError("Call simulate first.")

    #     if self.verbose:
    #         print("Disaggregating selected variables.")

    #     data_hourly_trans = self._fit_seasonal_hourly(refit=refit)

    #     # infer the target covariance and autocovariance structure
    #     # from the hourly untransformed data
    #     data_hourly = base.met_as_array(self.met, var_names=self.var_names)
    #     # the nans mess up the calculation of the covariance matrix
    #     data_hourly = data_hourly[:, np.all(np.isfinite(data_hourly), axis=0)]
    #     if isinstance(self.sum_interval, abc.Iterable):
    #         warnings.warn(
    #             "Per-variable sum_interval is not supported " "anymore."
    #         )
    #         disagg_len = self.sum_interval[0]
    #     else:
    #         disagg_len = self.sum_interval

    #     # disaggregate
    #     dht = np.array(data_hourly_trans)
    #     if "R" in self.var_names:
    #         if self.verbose:
    #             print("Transforming 'negative rain'")
    #         # rain_old = np.copy(dht[0])
    #         dht = self._negative_rain(
    #             dht,
    #             self.met["R"],
    #             self.data_doys_raw,
    #             # var_names=("theta", "ILWR", "u", "v")
    #         )

    #         # fig, axs = plt.subplots(self.K, sharex=True)
    #         # for val, ax in zip(dht, axs):
    #         #     ax.plot(dtimes, val)
    #         # axs[0].plot(dtimes, rain_old)
    #         # plt.show()
    #         sum_vars = self.var_names.index("R")
    #     else:
    #         sum_vars = None
    #     cov = np.cov(dht[:, np.all(np.isfinite(dht), axis=0)])
    #     self.data_hourly_trans = dht

    #     # self.plot_transformed_hourly()
    #     # plt.show()

    #     # generate the datetime information of the disaggregated time series.
    #     # how many timesteps are there per day in the input data?
    #     # we assume that the data is not finer than hours
    #     hours_unique = np.unique([dtime.hour for dtime in self.times_orig])
    #     # read timesteps per day (abbreviation, because it is used a lot as an
    #     # index modifier below
    #     tpd = len(hours_unique)
    #     self.dis_times = self._gen_sim_times(
    #         self.T_sim * tpd, output_resolution=1.0
    #     )

    #     def trans(data, doys):
    #         return seasonal_back(
    #             self.dist_sol,
    #             data,
    #             self.var_names,
    #             doys,
    #             solution_template="%s_hourly",
    #         )

    #     t_kwds = dict(doys=times.datetime2doy(self.dis_times))
    #     self.sim_dis = csim.disaggregate_piecewice(
    #         self.sim,
    #         autocov=data_hourly_trans,
    #         disagg_len=disagg_len,
    #         cov=cov,
    #         pool_size=25,
    #         trans=trans,
    #         t_kwds=t_kwds,
    #         verbose=True,
    #         sum_vars=sum_vars,
    #     )

    #     # retransform disaggregated variables
    #     sim_sea_dis = seasonal_back(
    #         self.dist_sol,
    #         self.sim_dis,
    #         self.var_names,
    #         doys=times.datetime2doy(self.dis_times),
    #         solution_template="%s_hourly",
    #     )
    #     # housekeeping
    #     self.sim_sea_dis = sim_sea_dis
    #     # some of the plotting routines depend on the availability of
    #     # the variable names we disaggregated
    #     self.var_names_dis = self.var_names
    #     return self.dis_times, sim_sea_dis

    def disaggregate(
        self,
        var_names_dis=None,
        event_dt=None,
        factors=None,
        doy_tolerance=15,
        latitude=None,
        longitude=None,
    ):
        """Disaggregate variables to hourly time steps by drawing from the
        residuals of the measured data

        Parameters
        ----------
        var_names_dis : list of str, optional
            names of variables to be disaggregated. if None all variables are
            disaggregated
        event_dt : array of datetimes, optional
            days on which factors will be applied to change daily cycle
            amplitudes.
        factors : sequence of length K
            see event_dt

        Returns
        -------
        times_out : array of datetime objects
        sim_sea_dis : array
            simulated variables in hourly time steps. the last day gets lost
        """
        if self.sim is None:
            raise RuntimeError("Call simulate first.")

        # disaggregate selected variables to hourly values, all the others
        # will be repeated 24 times a day. The last day gets lost
        if var_names_dis is None:
            if self.verbose:
                print("Disaggregating all variables")
            var_names_dis = self.var_names
        elif isinstance(var_names_dis, str):
            if self.verbose:
                print("Disaggregating selected variables.")
            var_names_dis = (var_names_dis,)

        if latitude is None:
            latitude = conf.latitude
        if longitude is None:
            longitude = conf.longitude

        # how many timesteps are there per day in the input data?
        # we assume that the data is not finer than hours
        hours_unique = np.unique([dtime.hour for dtime in self.times_orig])
        # read timesteps per day (abbreviation, because it is used a
        # lot as an index modifier below)
        tpd = len(hours_unique)

        self.dis_times = self._gen_sim_times(
            (self.T_sim - 1) * tpd, output_resolution=1.0
        )
        # reset cache
        self._dis_doys = None

        if factors is not None:
            factors = np.asarray(factors)[:, None]

        deltas_input, sim_interps = self._gen_deltas_input(
            var_names_dis,
            tpd,
            longitude=longitude,
            latitude=latitude,
        )
        deltas_drawn, sim_sea_dis = self._add_deltas(
            deltas_input,
            sim_interps,
            var_names_dis,
            tpd,
            event_dt=event_dt,
            factors=factors,
            doy_tolerance=doy_tolerance,
        )

        #  check for variables with limits
        for var_name in var_names_dis:
            var_i = self.var_names.index(var_name)
            limits = copy.copy(conf.par_known[var_name])
            u_kwds = dict()
            if var_name.startswith("Qsw"):

                def pot_s(doys, longitude, latitude):
                    hourly = pot_s_rad(
                        doys,
                        lat=latitude,
                        longt=longitude,
                        tz_mer=None,
                    )
                    return hourly * self.sum_interval[var_i]

                u_kwds = dict(longitude=longitude, latitude=latitude)
                limits["u"] = pot_s
            elif var_name.startswith("sun"):

                def sun_hours(doys):
                    dtimes = varwg.times.doy2datetime(doys)
                    sunrise, sunset = sunshine_riseset(
                        dtimes, longitude, latitude, tz_offset=None
                    )
                    doy_hours = (doys - doys.astype(int)) * 24
                    max_hours = np.full_like(doys, 60)
                    max_hours[(doy_hours < sunrise) & (doy_hours > sunset)] = 0
                    sunrise_dist = 60 * (sunrise - doy_hours)
                    sunset_dist = 60 * (sunset - doy_hours)
                    sunrise_mask = (sunrise_dist > 0) & (sunrise_dist < 60)
                    sunset_mask = (sunset_dist < 0) & (sunset_dist > -60)
                    max_hours[sunrise_mask] = sunrise_dist[sunrise_mask]
                    max_hours[sunset_mask] = 60 + sunset_dist[sunset_mask]
                    return max_hours * self.sum_interval[var_i]

                limits["u"] = sun_hours

            sim_interp = sim_interps[var_i]
            pos_mask = deltas_drawn[var_i] > 0
            neg_mask = ~pos_mask

            # we have to interpret a threshold as a lower limit (no negative
            # rain, please)
            seas_dist, _ = self.dist_sol[var_name]
            if hasattr(seas_dist, "dist") and hasattr(
                seas_dist.dist, "thresh"
            ):
                if limits is None:
                    limits = {}
                # limits["l"] = conf.array_gen(seas_dist.dist.thresh)
                limits["l"] = conf.array_gen(conf.threshold)
                # limits["l"] = conf.array_gen(0.)
                # limits["u"] = conf.array_gen(1.e12)

            if limits is not None and ("u" in limits or "uc" in limits):
                upper_func = limits["u"] if "u" in limits else limits["uc"]
                upper = (
                    upper_func(self.dis_doys[pos_mask], **u_kwds)
                    / self.sum_interval[var_i]
                )
                upper_perc = deltas_drawn[var_i, pos_mask]
                interps = sim_interp[pos_mask]
                upper_dis = interps + upper_perc * (upper - interps)
                sim_sea_dis[var_i, pos_mask] = upper_dis

            if limits is not None and ("l" in limits or "lc" in limits):
                lower_func = limits["l"] if "l" in limits else limits["lc"]
                lower = (
                    lower_func(self.dis_doys[neg_mask])
                    / self.sum_interval[var_i]
                )
                lower_perc = deltas_drawn[var_i, neg_mask]
                interps = sim_interp[neg_mask]
                lower_dis = interps + lower_perc * (interps - lower)
                lower_dis = np.where(lower_dis < lower, lower, lower_dis)
                sim_sea_dis[var_i, neg_mask] = lower_dis

        # sim_sea_dis = np.roll(sim_sea_dis, -1, axis=1)

        # fig, ax = plt.subplots()
        # ax.plot(sim_sea_dis[0], "-o", label="sim_sea_dis")
        # ax.plot(sim_interps[0], "-x", label="sim_interps")
        # daily_rain_out = self.sim_sea[self.var_names.index("R")]
        # m = self.T_sim * tpd - tpd
        # rain_mask_out = (daily_rain_out > 0).repeat(tpd)[:m]
        # for start_i, end_i in my.gaps(rain_mask_out):
        #     ax.axvspan(start_i, end_i, alpha=.5)
        # ax.legend(loc="best")
        # plt.show()

        self.var_names_dis = var_names_dis
        self.sim_sea_dis = sim_sea_dis

        if self.dump:
            # if varnames_dis==None, daily values are returned and dumped:
            self.outfilepath = dump_data(
                self.dis_times,
                sim_sea_dis,
                self.var_names,
                self.p,
                0 if self.q is None else self.q,
                "_disaggregated",
                out_dir=self.data_dir,
                conversions=conf.conversions,
            )

        return self.dis_times, sim_sea_dis

    def _prepare_output(self):
        """Makes the requested conversions defined in the config file and
        returns a dictionary for methods like to_dyresm or to_glm."""
        if self.sim_sea_dis is None:
            raise RuntimeError("Call disaggregate first.")

        dtimes, data, var_names = (
            self.dis_times,
            self.sim_sea_dis,
            self.var_names,
        )
        if conf.conversions:
            for conversion in list(conf.conversions):
                times, data, var_names = conversion(dtimes, data, var_names)
        met = {var_name: vals for var_name, vals in zip(var_names, data)}
        return dtimes, met

    def to_dyresm(
        self, outfilepath, ts=3600, info=None, wind_fct=1.0, output_rh=False
    ):
        """Converts output to a DYRESM input file.

        Parameters
        ----------
        outfilepath : str
            dyresm meteorological bc file
        ts : int
            timestep in dyresm-met-file in seconds, should be divisor of 86400
        info : information text string in file header
        wind_fct : float
            factor for wind speed. Default: 1.0 (for Konstanz wind data: 1.3)

        """
        times_, met_dict = self._prepare_output()

        # the now string! ready cut, easy to handle, simpson's individual
        # emperor stringettes - just the right length!
        now_str = datetime.datetime.now().isoformat(" ")
        moist_varname = "rh" if output_rh else "e"
        template_filepath = os.path.join(
            os.path.dirname(__file__), "dyresm_header_template"
        )
        with open(template_filepath) as template_file:
            template = template_file.read()
        dyresm_header = template.format(
            now_str=now_str, info=info, ts=ts, moist_varname=moist_varname
        )

        if not output_rh:
            if "e" not in met_dict:
                met_dict["e"] = meteox2y.rel2vap_p(
                    met_dict["rh"], met_dict["theta"]
                )
        if "U" not in met_dict:
            met_dict["U"] = avrwind.component2angle(
                met_dict["u"], met_dict["v"]
            )[1]
        met_dict["U"] *= wind_fct
        if "R" not in met_dict:
            met_dict["R"] = np.zeros_like(met_dict["theta"])
        data = np.array(
            [varwg.times.datetime2cwr(times_)]
            + [
                met_dict[var_name]
                for var_name in (
                    "Qsw",
                    "ILWR",
                    "theta",
                    "rh" if output_rh else "e",
                    "U",
                    "R",
                )
            ]
        ).T

        with open(outfilepath, "w") as outfile:
            outfile.write(dyresm_header)
        with open(outfilepath, "ab") as outfile:
            np.savetxt(
                outfile,
                data,
                fmt=(
                    b"%10.5f\t%7.3f\t%7.3f\t%6.2f\t"
                    + (b"%.3f" if output_rh else b"%4.1f")
                    + b"\t%5.2f\t%.6f"
                ),
            )

    def to_glm(self, outfilepath):
        """Converts output to a GLM input file.

        Parameters
        ----------
        outfilepath : path/filename
            glm meteorological bc file
        """
        dtimes, met = self._prepare_output()

        header = (
            "time,ShortWave,LongWave, AirTemp,RelHum, WindSpeed,Rain,"
            "Snow" + os.linesep
        )
        if "U" not in met:
            try:
                met["U"] = avrwind.component2angle(met["u"], met["v"])[1]
            except KeyError:
                warnings.warn("No wind information found. Filling in 0's!!")
                met["U"] = np.zeros_like(met[list(met.keys())[0]])
        if "R" not in met:
            warnings.warn(
                "No precipitation information available. Filling "
                " in with 0's!!"
            )
            met["R"] = np.zeros_like(met[list(met.keys())[0]])
        times_str = varwg.times.datetime2str(dtimes, "%Y-%m-%d %H:%M")
        lines = [
            ",".join(
                [time_str]
                + [
                    "%.6f" % met[key][i]
                    for key in ("Qsw", "ILWR", "theta", "rh", "U", "R")
                ]
                + ["0.0" + os.linesep]  # snow
            )
            for i, time_str in enumerate(times_str)
        ]
        with open(outfilepath, "w") as glm_file:
            glm_file.write(header)
            glm_file.writelines(lines)

    def to_gotm(self, outfilepath, sw_outfilepath=None, rain_outfilepath=None):
        """Converts output to a GOTM input file set (met and separate
        short-wave file).

        Parameters
        ----------
        outfilepath : path/filename
            GOTM meteorological bc file
        sw_outfilepath : path/filename
            GOTM short-wave bc file
        rain_outfilepath : path/filename
            GOTM precipitation bc file
        """
        if sw_outfilepath is None:
            sw_outfilepath = outfilepath[: -len(".dat")] + "_swr.dat"
        if rain_outfilepath is None:
            rain_outfilepath = outfilepath[: -len(".dat")] + "_precip_file.dat"
        dtimes, met = self._prepare_output()

        if "U" not in met:
            try:
                met["wdir"], met["U"] = avrwind.component2angle(
                    met["u"], met["v"]
                )
                # gideon: "I think the cloud cover and wind direction
                # should both be 0 (for the time being...)"
                met["wdir"][:] = 0
            except KeyError:
                warnings.warn("No wind information found. Filling in 0's!!")
                met["wdir"] = met["U"] = np.zeros_like(
                    met[list(met.keys())[0]]
                )
        if "R" not in met:
            warnings.warn(
                "No precipitation information available. Filling "
                " in with 0's!!"
            )
            met["R"] = np.zeros_like(met[list(met.keys())[0]])
        if "air_pressure" not in met:
            met["air_pressure"] = np.full_like(met[list(met.keys())[0]], 1000)
        if "cloud_cover" not in met:
            ilwr = meteox2y.temp2lw(met["theta"])
            met["cloud_cover"] = meteox2y.lw2clouds(
                ilwr, met["theta"], rh=met["rh"]
            )
            met["cloud_cover"][:] = 0

        # relative humidity has to be in %
        rh = met["rh"]
        if np.max(rh) <= 1.5:
            met["rh"] *= 100

        times_str = varwg.times.datetime2str(dtimes, "%Y-%m-%d %H:%M:%S")
        var_names = "U", "cloud_cover", "air_pressure", "theta", "rh", "wdir"
        # fmts = "%.2f", "%.6f", "%.0f", "%.3f", "%.6f"
        lines = [
            "\t".join(
                [time_str]
                + [conf.out_format[key] % met[key][i] for key in var_names]
                + [os.linesep]
            )
            for i, time_str in enumerate(times_str)
        ]

        with open(outfilepath, "w") as gotm_file:
            # gotm_file.write(header)
            gotm_file.writelines(lines)

        # separate short-wave file
        lines = [
            "%s\t%s%s"
            % (time_str, conf.out_format["Qsw"] % met["Qsw"][i], os.linesep)
            for i, time_str in enumerate(times_str)
        ]
        with open(sw_outfilepath, "w") as gotm_file:
            gotm_file.writelines(lines)

        # separate precipitation file
        lines = [
            "%s\t%s%s"
            % (time_str, conf.out_format["R"] % met["R"][i], os.linesep)
            for i, time_str in enumerate(times_str)
        ]
        with open(rain_outfilepath, "w") as gotm_file:
            gotm_file.writelines(lines)

    def to_df(
        self, kind="hourly input", var_names=None, *, with_conversions=True
    ):
        if var_names is None:
            var_names = self.var_names

        data = None
        if kind == "hourly input":
            if "e" in var_names and "e" not in self.met:
                self.met["e"] = meteox2y.rel2vap_p(
                    self.met["rh"], self.met["theta"]
                )
            data = np.array([self.met[var_name] for var_name in var_names])
            index = self.times_orig
        elif kind == "daily input":
            data = self.data_raw / self.sum_interval
            index = self.times
        elif kind == "daily input trans":
            data = self.data_trans
            index = self.times
        elif kind == "daily output trans":
            data = self.sim
            index = self.sim_times
        elif kind == "daily output":
            data = self.sim_sea
            index = self.sim_times
        elif kind == "hourly output":
            data = self.sim_sea_dis
            index = self.dis_times
        else:
            raise RuntimeError(f"kind={kind} not understood")
            return
        if data is None and "output" in kind:
            raise RuntimeError("Call simulate, before requesting output")
        if data.shape[0] != len(var_names):
            data = np.array(
                [data[self.var_names.index(name)] for name in var_names]
            )
        # do not trust that the conversions have any side-effects on
        # its parameters!
        index, data = map(np.copy, (index, data))
        var_names = [name for name in var_names]
        if with_conversions:
            if (
                "trans" not in kind
                and "input" not in kind
                and conf.conversions
            ):
                for conversion in conf.conversions:
                    index, data, var_names = conversion(index, data, var_names)
        df = pd.DataFrame(data=data.T, columns=var_names, index=index)
        df.name = kind
        df.index.name = "time"
        return df

    def infill_trans_nans(self):
        if self.p is None:
            if self.verbose:
                print("Fitting seasonal VAR for infilling.")
            self.fit(**self.fit_kwds)
        data_infilled = models.SVAR_LS_fill(
            self.Bs, self.sigma_us, self.data_doys, self.data_trans
        )
        n_nan_timesteps = (np.isnan(self.data_trans).sum(axis=0) > 0).sum()
        if self.verbose:
            nan_perc = n_nan_timesteps / self.T_summed * 100
            if n_nan_timesteps:
                print(
                    f"Filled in {n_nan_timesteps} time steps ({nan_perc:.1f}%)"
                )
            for var_i, values in enumerate(self.data_trans):
                n_nan_timesteps = np.isnan(values).sum()
                if n_nan_timesteps:
                    nan_perc = n_nan_timesteps / self.T_summed * 100
                    var_name = self.var_names[var_i]
                    print(f"\t{var_name}: {n_nan_timesteps} ({nan_perc:.1f}%)")

        # fig, axs = plt.subplots(nrows=self.K, ncols=1, sharex=True)
        # for var_i, var_name in enumerate(self.var_names):
        #     ax = axs[var_i]
        #     ax.plot(self.times, self.data_trans[var_i], color="k")
        #     ax.plot(self.times, data_infilled[var_i], "--", color="b")
        #     ax.set_title(var_name)
        # fig.suptitle(self.station_name)

        # fig, axs = self.plot_meteogram_hourly()
        # fig, axs = self.plot_meteogram_trans()

        self.data_trans = data_infilled

        # fig, axs = self.plot_meteogram_trans(figs=fig, axss=axs)
        # fig, axs = self.plot_meteogram_daily()

        self.data_raw = seasonal_back(
            self.dist_sol, self.data_trans, self.var_names, self.data_doys
        )

        if n_nan_timesteps:
            self.fit(**self.fit_kwds)
        # fig, axs = self.plot_meteogram_daily(figs=fig, axss=axs)
        # plt.show()

        return data_infilled

    def predict(self, dtimes=None, data_past=None, T=1, n_realizations=1):
        """Predict the next T steps given the past data.

        Parameters
        ----------
        dtimes : None or (p,) ndarray, optional
            datetimes of past timesteps, if None, the last p datetimes from
            the input data are used
        data_past : None or (K, p)) ndarray, optional
            Past on which to predict the future. p is the order of the VAR -
            process (it is stored in self.p after self.fit() is called).
            If None, data from the input is used.
        T : int, optional
            number of timesteps to predict.
        n_realizations : int, optional
            number of realizations of prediction

        Returns
        -------
        prediction : (K, T) or (K, T, n_realizations) ndarray
        """
        if self.p is None:
            self.fit(**self.fit_kwds)
        if dtimes is None:
            dtimes = self.times[-self.p :]
        if data_past is None:
            data_past = self.data_raw[:, -self.p :]
        past_doys = varwg.times.datetime2doy(dtimes)
        data_past_trans = self._fit_seasonal(values=data_past, doys=past_doys)[
            0
        ]
        prediction = models.VAR_LS_predict(
            data_past_trans, self.AM, self.sigma_u, T, n_realizations
        )
        prediction = prediction.reshape((-1, T, n_realizations))

        # calculate the output resolution in days
        first_doy, second_doy = varwg.times.datetime2doy(self.times[:2])
        delta_doy = second_doy - first_doy
        prediction_doys = past_doys[-1] + delta_doy * np.arange(1, T + 1)
        prediction_dtimes = varwg.times.doy2datetime(
            prediction_doys, year=dtimes[-1].year
        )
        prediction_doys = np.where(
            prediction_doys > 366, prediction_doys - 366, prediction_doys
        )
        prediction_sea = np.atleast_3d(np.empty_like(prediction))

        for r in range(n_realizations):
            prediction_sea[..., r] = seasonal_back(
                self.dist_sol,
                prediction[..., r],
                self.var_names,
                doys=prediction_doys,
            )
        prediction_sea /= self.sum_interval[:, np.newaxis, :]
        return prediction_dtimes, np.squeeze(prediction_sea)

    def plot_prediction(self, dtimes=None, data_past=None, T=1, hindcast=True):
        # don't be verbose for this part (we fetch previous fits all the time
        # which is rather boring)
        verbose = self.verbose
        self.verbose = False
        if self.p is None:
            self.fit(**self.fit_kwds)
        n_prev_steps = (4 if hindcast else 2) * self.p
        if dtimes is None:
            dtimes = self.times[-n_prev_steps:]
        if data_past is None:
            data_past = self.data_raw[:, -n_prev_steps:]

        prediction_dtimes, prediction = self.predict(
            dtimes[-self.p :], data_past[:, -self.p :], T
        )

        fig, axes = plt.subplots(nrows=self.K, sharex=True, squeeze=True)
        fig.suptitle(
            "Prediction made on %s"
            % varwg.times.datetime2str(varwg.times.datetime.now())
        )
        for var_i, (ax, var_name) in enumerate(zip(axes, self.var_names)):
            ax.plot(
                dtimes[self.p :],
                data_past[var_i, self.p :] / self.sum_interval[var_i],
                "k",
                linewidth=2,
            )
            ax.plot(prediction_dtimes, prediction[var_i], "b-x", linewidth=2)
            ax.grid()
            ax.set_ylabel(r"%s %s" % (var_name, conf.units[var_name]))
            ax.set_title(conf.long_var_names[var_name])

        if hindcast:
            for shift in range(self.p + 1, n_prev_steps):
                hindcast_dtimes, hindcast = self.predict(
                    dtimes[:shift],
                    data_past[:, :shift],
                    n_prev_steps - shift,
                    1,
                )
                col = plt.cm.jet(
                    1
                    - float(shift - self.p - 1) / (n_prev_steps - self.p - 1),
                    alpha=0.5,
                )
                for var_i, ax in enumerate(axes):
                    ax.plot(
                        hindcast_dtimes,
                        hindcast[var_i],
                        "-",
                        color=col,
                        linewidth=1,
                    )
                    # connect the hindcast with the end of its supporting data
                    ax.plot(
                        [dtimes[shift - 1], hindcast_dtimes[0]],
                        [
                            data_past[var_i, shift - 1]
                            / self.sum_interval[var_i],
                            np.atleast_1d(hindcast[var_i])[0],
                        ],
                        "-x",
                        color=col,
                        linewidth=1,
                    )
        ti = plt.xticks()[0]
        la_ = [
            varwg.times.datetime2str(
                varwg.times.ordinal2datetime(tii), "%d.%m. %H h"
            )
            for tii in ti
        ]
        plt.xticks(ti, la_, rotation=30)

        self.verbose = verbose

    def _gen_m_trend(
        self, prim_i, prim_index, prim_is_normal, sigma, scale, scale_nn
    ):
        theta_grad = self.theta_grad[prim_i]
        if np.isnan(theta_grad):
            self._m_trend_single += [np.zeros(self.K)]
            return self._m_trend_single[-1]
        # e.g. gradual temperature change
        primvar_trend = theta_grad * self.sum_interval[prim_index]
        # looking for the mean of the standard deviation
        if prim_is_normal:
            m_trend_primvar = primvar_trend / np.mean(sigma)
        else:
            # TODO: the following is most likely wrong!!!
            m_trend_primvar = np.mean(scale_nn(primvar_trend))
        self._m_trend_single += [m_trend_primvar * scale]
        return self._m_trend_single[-1]

    def _gen_m(
        self, prim_i, prim_index, prim_is_normal, sigma, scale, scale_nn
    ):
        if self.theta_incr is not None:
            theta_incr = self.theta_incr[prim_i]
            if np.isnan(theta_incr):
                self._m_single += [np.zeros((self.K, self.T_sim))]
                return self._m_single[-1]

        intercept = (
            # self.data_trans.mean(axis=1)
            np.nanmean(self.data_trans, axis=1)
            - scale * self.data_trans[prim_index].mean()
        ).reshape((self.K, 1))
        intercept[prim_index] = 0
        if self.theta_incr is not None:
            # change in mean in the real world
            delta_primvar = theta_incr * self.sum_interval[prim_index]
            if prim_is_normal:
                m_primvar = delta_primvar / sigma
            else:
                m_primvar = scale_nn(delta_primvar)
        else:
            m_primvar = 0
        if self.phase_randomize and self.phase_randomize_vary_mean:
            var_mean_scale = (
                self.phase_randomize_vary_mean
                if isinstance(self.phase_randomize_vary_mean, float)
                else 0.5
            )
            m_primvar += var_mean_scale * varwg.rng.normal()
            return m_primvar

        else:
            self._m_single += [
                m_primvar * scale.reshape((self.K, 1)) + intercept
            ]
            return self._m_single[-1]

    def _gen_m_t(
        self,
        prim_i,
        prim_index,
        prim_is_normal,
        sigma,
        scale,
        scale_nn,
        scale_nn_simple,
        _T,
    ):
        m_t = np.zeros((self.K, self.T_sim))
        # das war an der Tafel im Seminarraum 2:
        # mean_arrival: mittlere Zeit zwischen Aenderungen (kalte/warme
        # Perioden)
        # disturbance_std: Stabw der disturbances
        if self.disturbance_std is not None:
            disturbance_std = self.disturbance_std[prim_i]
            mean_arrival = self.mean_arrival[prim_i]
            disturbance_std = disturbance_std * self.sum_interval[prim_index]
            if prim_is_normal:
                disturbance_std /= np.mean(sigma)
            else:
                disturbance_std = np.mean(scale_nn(disturbance_std))
            m_t_primvar = interannual_variability(
                self.T_sim, mean_arrival, disturbance_std
            )
            m_t += m_t_primvar * scale.reshape((self.K, 1))

        if (
            self.climate_signal is not None
            and self.climate_signal[prim_i] is not None
        ):
            climate_diff = self._get_climate_diff(_T, prim_i)
            if prim_is_normal:
                climate_diff /= sigma
            else:
                climate_diff = scale_nn_simple(climate_diff)
            m_t += climate_diff * scale.reshape((self.K, 1))
        self._m_t_single += [m_t]
        return m_t

    @property
    def cov_trans(self):
        """The covariance matrix of the transformed data.
        If seasonal is True, this is of (365, K, K)-shape.
        """
        if self._cov_trans is None:
            if self.seasonal:
                covs = []
                for doy_ii, doy in enumerate(self.unique_doys):
                    ii = (self.data_doys > doy - self.doy_width) & (
                        self.data_doys <= doy + self.doy_width
                    )
                    if (doy - self.doy_width) < 0:
                        ii |= self.doys > (365.0 - self.doy_width + doy)
                    if (doy + self.doy_width) > 365:
                        ii |= self.doys < (doy + self.doy_width - 365.0)
                    covs += np.cov(self.data_trans[ii])
                covs = [
                    [
                        my.fourier_approx(covs[:, i, j], 4)
                        for i in range(self.K)
                    ]
                    for j in range(self.K)
                ]
                self._cov_trans = np.array(covs)
            else:
                self._cov_trans = np.cov(self.data_trans)
        return self._cov_trans

    def _scale_nn(self, x, seas_dist, trig_params):
        return my.pickle_cache(
            str(
                Path(self.seasonal_cache_file).parent
                / f"qq_shift_{self.station_name}_{x}_%s.pkl"
            ),
            warn=False,
            # clear_cache=True,
        )(seas_dist.qq_shift)(
            x,
            trig_params,
            # doys=self.sim_doys
        )

    def _scenario_parameters(
        self,
        theta_incr=None,
        theta_grad=None,
        disturbance_std=None,
        climate_signal=None,
        phase_randomize=False,
        scale_prim=True,
        primary_var_sim=None,
    ):
        """Prepares the scenario parameters, m and m_trend for the
        std.-normal world.

        Parameters
        ----------
        phase_randomize : boolean, optional
            Use phase randomization for VAR-residuals
        scale_prim : boolean, optional
            Allow primary variables to influence each other.
            Only effective when using multiple primary variables!
        """

        def str2tuple(thing):
            if isinstance(thing, str):
                return (thing,)
            return thing

        if primary_var_sim is None:
            prim_vars = str2tuple(self.primary_var)
            primary_var_ii = self.primary_var_ii
        else:
            prim_vars = str2tuple(primary_var_sim)
            primary_var_ii = [
                self.var_names.index(prim_var) for prim_var in prim_vars
            ]

        m = np.zeros((self.K, self.T_sim))
        m_t = np.zeros_like(m)
        m_trend = np.zeros(self.K)
        ScenParameters = namedtuple(
            "scenario_parameters", ("m", "m_t", "m_trend")
        )

        for prim_i, prim in enumerate(prim_vars):
            # prim_i is index inside sequence of primary variables.
            # prim_index is index inside sequence of all variables.
            # prim_index = self.primary_var_ii[prim_i]
            prim_index = primary_var_ii[prim_i]
            # scale other variables according to linear regression
            finite_row_mask = np.all(np.isfinite(self.data_trans), axis=0)
            data_trans = self.data_trans[:, finite_row_mask]
            scale = np.cov(data_trans)[prim_index] / np.var(
                data_trans[prim_index]
            )

            # do not change the primary variable by itself
            scale[prim_index] = 1
            if not scale_prim:
                # do not change the other primary variables
                scale[self.primary_var_ii] = 1
            if prim in self.var_names:
                seas_dist, trig_params = self.dist_sol[prim]
                if hasattr(seas_dist, "dist"):
                    prim_is_normal = isinstance(
                        seas_dist.dist, distributions.Normal
                    )
                else:
                    prim_is_normal = False

                if not prim_is_normal:
                    scale_nn = functools.partial(
                        self._scale_nn,
                        seas_dist=seas_dist,
                        trig_params=trig_params,
                    )

                    # means = seas_dist.mean(trig_params, self.sim_doys)

                    # we might exist in a loop and want to avoid
                    # calculating the medians all the time
                    if not hasattr(seas_dist, "medians"):
                        seas_dist.medians = seas_dist.median(
                            trig_params, self.sim_doys
                        )

                    def scale_nn_simple(x):
                        if seas_dist.medians.shape != x.shape:
                            seas_dist.medians = seas_dist.median(
                                trig_params, self.sim_doys
                            )
                        return distributions.norm.ppf(
                            seas_dist.cdf(
                                trig_params,
                                seas_dist.medians + x,
                                self.sim_doys,
                            )
                        )

                else:
                    scale_nn = None
                    scale_nn_simple = None

                # get the doy-specific distribution parameters
                _T = (2 * np.pi / 365 * self.sim_doys)[np.newaxis, :]
                if prim_is_normal:
                    sigma = seas_dist.trig2pars(trig_params, _T)[1]
                else:
                    sigma = None

                if theta_grad is not None:
                    m_trend += self._gen_m_trend(
                        prim_i,
                        prim_index,
                        prim_is_normal,
                        sigma,
                        scale,
                        scale_nn,
                    )
                else:
                    self._m_trend_single += [np.zeros(self.K)]

                if theta_incr is not None or self.phase_randomize_vary_mean:
                    m += self._gen_m(
                        prim_i,
                        prim_index,
                        prim_is_normal,
                        sigma,
                        scale,
                        scale_nn,
                    )
                else:
                    self._m_single += [np.zeros((self.K, self.T_sim))]

                if disturbance_std is not None or climate_signal is not None:
                    m_t += self._gen_m_t(
                        prim_i,
                        prim_index,
                        prim_is_normal,
                        sigma,
                        scale,
                        scale_nn,
                        scale_nn_simple,
                        _T,
                    )

        return ScenParameters(m, m_t, m_trend)

    def _get_climate_diff(self, _T, prim_i=0):
        """Calculates the difference between the seasonal mean of the primary
        variable and the given climate signal."""
        var_name = self.primary_var[prim_i]
        return (
            # np.atleast_2d(self.climate_signal)[prim_i]
            self.climate_signal[prim_i]
            - self.fitted_medians(var_name)
        ) * self.sum_interval_dict[var_name]

    def _location_shift_normal(self, sim):
        for prim_i, prim in enumerate(self.primary_var):
            prim_index = self.var_names.index(prim)
            dummy_time = np.arange(self.T_sim, dtype=float)
            m = self._m_single[prim_i][prim_index]
            m_t = mt[prim_i][prim_index] if (mt := self._m_t_single) else 0
            m_trend = self._m_trend_single[prim_i][prim_index]
            sim[prim_index] -= m + m_t + dummy_time / self.T_sim * m_trend
        return sim

    def _location_shift_back(self, sim_sea):
        for prim_i, prim in enumerate(self.primary_var):
            prim_index = self.var_names.index(prim)

            if self.theta_incr is not None:
                theta_incr = self.theta_incr[prim_i]
                if not np.isnan(theta_incr):
                    sim_sea[prim_index] += theta_incr

            if self.theta_grad is not None:
                theta_grad = self.theta_grad[prim_i]
                if not np.isnan(theta_grad):
                    dummy_time = np.arange(self.T_sim, dtype=float)
                    sim_sea[prim_index] += dummy_time / self.T_sim * theta_grad

            if self.climate_signal is not None:
                if self.climate_signal[prim_i] is not None:
                    # get the doy-specific distribution parameters
                    _T = (2 * np.pi / 365 * self.sim_doys)[np.newaxis, :]
                    climate_diff = self._get_climate_diff(_T, prim_i)
                    sim_sea[prim_index] += (
                        climate_diff / self.sum_interval[prim_index]
                    )
        return sim_sea

    def random_dryness(
        self,
        T=None,
        start_str=None,
        stop_str=None,
        n_events=None,
        duration_min=2,
        duration_max=7,
        month_start=6,
        month_end=9,
        event_dryness=0.4,
    ):
        """This is a very special method designed to generate dry hot events
        for Lake Kinneret.

        Parameters
        ----------
        T : None or int, optional
            Number of timesteps to simulate. T=None produces as many simulation
            timesteps as there were in the data used for fitting the VARMA
            process.
        start_str : str of format '%m-%d %H', optional
            String representation of the start date of the simulation.
        stop_str : str of format '%m-%d %H', optional
            String representation of the end date of the simulation.
        n_events : int or None
            Number of events to generate. If none there will be an average of 3
            events per year.
        duration_min : int
            Minimum length of events
        duration_max : int
            Maximum length of events
        month_start : int
            Month when the events can start to happen (january = 1)
        month_end : int
            Latest month in which the events can happen (january = 1)
        event_dryness : float
            The value of relative humidity that will be set during the events.

        Returns
        -------
        rh_signal : ndarray, dtype=float
            To be used as climate_signal in VG.simulate
        event_dt : ndarray, dtype=object (datetime)
            Datetimes of the events. Can be passed to VG.disaggregate later.
        """
        if T is None:
            self.T_sim = self.T_summed
        else:
            self.T_sim = T
        self.sim_times = self._gen_sim_times(T, start_str, stop_str)
        if n_events is None:
            n_events = 3 * int(len(self.sim_doys) / 365.0)

        rh_signal = np.copy(self.fitted_medians("rh", self.sim_doys))
        months = varwg.times.time_part(self.sim_times, "%m")
        summer_ii = np.where((months >= month_start) & (months <= month_end))[
            0
        ]
        durations = varwg.rng.integers(duration_min, duration_max + 1, n_events)
        for event_i in range(n_events):
            i = varwg.rng.choice(summer_ii)
            duration = durations[event_i]
            rh_signal[i : i + duration] = event_dryness

        event_mask = np.where(rh_signal <= event_dryness)[0]
        event_dt = self.sim_times[event_mask]

        return rh_signal, event_dt


if __name__ == "__main__":
    # warnings.simplefilter("error", RuntimeWarning)

    # import config_konstanz_disag as conf
    import config_konstanz as conf

    from varwg.core import plotting

    base.conf = plotting.conf = conf
    met_vg = VG(
        ("R", "theta", "ILWR", "Qsw", "rh", "u", "v"),
        # non_rain=("theta", "Qsw", "rh"),
        rain_method="regression",
        # rain_method="distance",
        neg_rain_doy_width=30,
        neg_rain_fft_order=2,
        # refit=True,
        # refit="R",
        verbose=True,
        dump_data=False,
        # plot=True
    )

    # met_vg.fit(seasonal=True)
    # # met_vg.fit(seasonal=False)
    # simt, sim = met_vg.simulate(T=100*365, phase_randomize=True)
    # # met_vg.disaggregate()
    # # simt, sim = met_vg.simulate(start_str="01.01.2000 00:00:00",
    # #                             stop_str="31.12.3000 00:00:00",
    # #                             resample=False)
    # # simt_dis, sim_dis = met_vg.disaggregate()
    # # met_vg.plot_all()
    # met_vg.plot_meteogramm_trans()
    # met_vg.plot_exceedance_daily()
    # # met_vg.plot_qq()
    # met_vg.plot_daily_fit("R")
    # met_vg.plot_monthly_hists("R")
    # fig, ax = plt.subplots(nrows=1, ncols=1)
    # rr, sol = met_vg.dist_sol["R"]
    # params = rr.all_parameters_dict(sol, doys=np.arange(365))
    # ax.plot(params["q_thresh"])
    # plt.show()



"""Pytest fixtures for VarWG tests."""

from pathlib import Path
import tempfile

import numpy as np
import pytest

import varwg

# Test configuration
seed = 0
p = 3
fit_kwds = dict(p=p, fft_order=3, doy_width=15, seasonal=True)
var_names = (
    "R",
    "theta",
    "Qsw",
    "ILWR",
    "rh",
    "u",
    "v",
)

# File paths
script_home = Path(varwg.__file__).parent
met_file = script_home / "sample.met"
data_dir = Path(varwg.core.__file__).parent / "tests" / "data"
sim_file = data_dir / "test_out_sample.met"


@pytest.fixture(scope="session")
def test_data_dir():
    """Create a session-wide temporary directory for test data."""
    tmpdir = tempfile.mkdtemp("vg_test_session")
    yield tmpdir
    # Note: Not cleaning up for now to allow inspection
    # import shutil
    # shutil.rmtree(tmpdir)


@pytest.fixture(scope="session")
def sample_sim():
    """Load sample simulation data once per session."""
    met = varwg.read_met(sim_file, verbose=False, with_conversions=True)[1]
    return np.array([met[var_name] for var_name in var_names])


@pytest.fixture(scope="session")
def varwg_kwds(test_data_dir):
    """Common VarWG initialization keywords."""
    return dict(
        refit=True,
        data_dir=test_data_dir,
        cache_dir=test_data_dir,
        met_file=met_file,
        verbose=False,
        infill=True,
        station_name="test",
    )


@pytest.fixture(scope="session")
def varwg_regr(varwg_kwds):
    """Session-scoped VarWG instance with regression rain method."""
    varwg.reseed(seed)
    met_varwg = varwg.VarWG(var_names, rain_method="regression", **varwg_kwds)
    met_varwg.fit(**fit_kwds)
    return met_varwg


@pytest.fixture(scope="session")
def varwg_dist(varwg_kwds):
    """Session-scoped VarWG instance with distance rain method."""
    met_varwg = varwg.VarWG(var_names, rain_method="distance", **varwg_kwds)
    met_varwg.fit(**fit_kwds)
    return met_varwg


@pytest.fixture(scope="session")
def varwg_sim(varwg_kwds):
    """Session-scoped VarWG instance with simulation rain method."""
    met_varwg = varwg.VarWG(var_names, rain_method="simulation", **varwg_kwds)
    met_varwg.fit(**fit_kwds)
    return met_varwg


@pytest.fixture
def varwg_regr_fresh(varwg_kwds):
    """Function-scoped VarWG instance for tests that need to call fit() with different params."""
    varwg.reseed(seed)
    met_varwg = varwg.VarWG(var_names, rain_method="regression", **varwg_kwds)
    met_varwg.fit(**fit_kwds)
    return met_varwg



import bottleneck
import numpy as np
from scipy import stats
import xarray as xr


def brunner_compound(sti, spi, sequential=False, progress=False):
    """Rank-based hot-dry index.


    Notes
    -----
    Brunner 2021 uses E-GPD for precipitation and a STI index for
    temperature. This implementation just uses empirical ranks.

    References
    ----------
    Brunner, Manuela I., Eric Gilleland, and Andrew W. Wood.
    “Space–Time Dependence of Compound Hot–Dry Events in the United
    States: Assessment Using a Multi-Site Multi-Variable Weather
    Generator.” Earth System Dynamics 12, no. 2 (May 19, 2021):
    621–34. https://doi.org/10.5194/esd-12-621-2021.
    """
    if progress:
        from tqdm import tqdm as progress
    else:

        def progress(x, *args, **kwds):
            return x

    assert len(sti) == len(spi)
    sti_ranks = bottleneck.nanrankdata(sti, axis=-1)
    spi_ranks = bottleneck.nanrankdata(-spi, axis=-1)
    try:
        if sequential:
            raise MemoryError
        bc = np.mean(
            (sti_ranks[..., None, :] <= sti_ranks[..., :, None])
            & (spi_ranks[..., None, :] <= spi_ranks[..., :, None]),
            axis=-1,
        )
        nan_mask = np.isnan(sti_ranks) | np.isnan(spi_ranks)
        bc[nan_mask] = np.nan
    except MemoryError:
        shape_before = sti_ranks.shape
        sti_ranks = sti_ranks.reshape(-1, shape_before[-1])
        spi_ranks = spi_ranks.reshape(-1, shape_before[-1])
        bc = np.full_like(sti_ranks, np.nan)
        for row_i in progress(range(len(sti_ranks)), total=len(sti_ranks)):
            for i, (sti_rank, spi_rank) in enumerate(
                zip(sti_ranks[row_i], spi_ranks[row_i])
            ):
                if np.isnan(sti_rank) or np.isnan(spi_rank):
                    continue
                sti_row = sti_ranks[row_i]
                spi_row = spi_ranks[row_i]
                finite_mask = np.isfinite(sti_row) & np.isfinite(spi_row)
                if np.any(finite_mask):
                    bc[row_i, i] = np.mean(
                        (sti_row <= sti_rank) & (spi_row <= spi_rank),
                        # (sti_row[finite_mask] <= sti_rank)
                        # & (spi_row[finite_mask] <= spi_rank),
                        axis=-1,
                    )
        bc = bc.reshape(shape_before)
    return bc


def _measure_ar(
    varname, distribution, agg_func_name, obs_ar, weeks=1, reference=None
):
    obs_ar = obs_ar.transpose(..., "time").rolling(time=weeks * 7)
    obs_ar = getattr(obs_ar, agg_func_name)()
    if reference is None:
        reference = obs_ar
    else:
        reference = reference.rolling(time=weeks * 7)
        reference = getattr(reference, agg_func_name)()
    months = reference.time.dt.month
    param_names = [param.name for param in distribution._param_info()]
    param_coords = {key: obs_ar.coords[key] for key in obs_ar.dims}
    param_coords["param"] = param_names
    params = xr.DataArray(
        np.empty(obs_ar.shape + (len(param_names),)),
        coords=param_coords,
    )
    for month, monthly_data in reference.groupby(months):
        flat = monthly_data.data.reshape(-1, monthly_data.sizes["time"])
        params_flat = np.array(
            [distribution.fit(row[np.isfinite(row)]) for row in flat]
        )
        params.loc[dict(time=(months == month))] = params_flat.reshape(
            monthly_data.shape[:-1] + (1, len(param_names))
        )
    measure_data = stats.norm.ppf(
        distribution.cdf(
            obs_ar,
            **{
                key: val.data
                for key, val in params.to_dataset("param").items()
            },
        )
    )
    measure_coords = obs_ar.coords.copy()
    measure_ar = xr.DataArray(
        measure_data, coords=measure_coords, name=varname
    )
    return measure_ar


def STI_ds(obs_ds, weeks=1, reference=None):
    sti_ar = _measure_ar(
        "sti",
        stats.norm,
        "mean",
        obs_ds.to_dataarray("variable"),
        weeks=weeks,
        reference=reference,
    )
    return xr.Dataset(dict(sti=sti_ar))


def STI_ar(obs_ar, weeks=1, reference=None):
    return _measure_ar(
        "sti",
        stats.norm,
        "mean",
        obs_ar,
        weeks=weeks,
        reference=reference,
    )


def SPI_ds(obs_ds, weeks=1, reference=None):
    spi_ar = _measure_ar(
        "spi",
        stats.genpareto,
        "sum",
        obs_ds.to_dataarray("variable"),
        weeks=weeks,
        reference=reference,
    )
    return xr.Dataset(dict(spi=spi_ar))


def SPI_ar(obs_ar, weeks=1, reference=None):
    return _measure_ar(
        "spi",
        stats.genpareto,
        "sum",
        obs_ar,
        weeks=weeks,
        reference=reference,
    )


if __name__ == "__main__":
    import dwd_opendata

    theta = dwd_opendata.load_data(
        ("Stötten", "Konstanz"), "air_temperature", time="hourly"
    ).to_dataset("station")
    theta = (
        theta.interpolate_na("time")
        .transpose(..., "time")
        .squeeze()
        .resample(time="1d")
        .sum()
        .sel(time=slice("1990", "2020"))
        .dropna("time")
    )
    sti = STI_ds(theta, weeks=6)

    prec = dwd_opendata.load_data(
        ("Stötten", "Konstanz"), "precipitation", time="hourly"
    ).to_dataset("station")
    prec = (
        prec.interpolate_na("time")
        .transpose(..., "time")
        .squeeze()
        .resample(time="1d")
        .sum()
        .sel(time=slice("1990", "2020"))
        .dropna("time")
    )
    spi = SPI_ds(prec, weeks=6)



"""Generates data/test_out_sample.met using ../sample.met as input."""

import shutil
import tempfile

import varwg
from pathlib import Path
from varwg.core.tests import test_core as test_vg

config_template = varwg.config_template
varwg.set_conf(config_template)

T = test_vg.T
fit_kwds = test_vg.fit_kwds


def main():
    data_filepath = Path(varwg.core.__file__).parent / "tests" / "data"
    test_in_data_filepath = data_filepath / "sample.met"
    if not test_in_data_filepath.exists():
        # try to look where the stand-alone vg has its sample data
        test_in_data_filepath = (
            Path("..") / Path(varwg.__file__).parent / "sample.met"
        )
    test_out_data_filepath = data_filepath / "test_out_sample.met"
    # in order not to refit the present fit...
    cache_dir = tempfile.mkdtemp("vg_test_data_gen")

    varwg.reseed(test_vg.seed)
    met_vg = varwg.VG(
        test_vg.var_names,
        met_file=test_in_data_filepath,
        cache_dir=cache_dir,
        data_dir=cache_dir,
        refit=True,
        verbose=True,
        infill=True,
        rain_method="regression",
    )
    met_vg.fit(**fit_kwds)
    varwg.reseed(test_vg.seed)
    met_vg.simulate(T=T)
    sim = met_vg.disaggregate(test_vg.disagg_varnames)[1]
    met_vg.to_df("hourly output", with_conversions=True).to_csv(
        test_out_data_filepath, sep="\t"
    )

    # do a roundtrip test
    met = varwg.read_met(
        test_out_data_filepath, verbose=True, with_conversions=True
    )[1]
    import numpy as np
    import numpy.testing as npt

    sample_sim = np.array([met[var_name] for var_name in test_vg.var_names])
    npt.assert_almost_equal(sim, sample_sim)

    shutil.rmtree(cache_dir)

    return met_vg


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    met_vg = main()
    met_vg.plot_meteogram_daily()
    plt.show()



"""
Created on 26.02.2014

@author: dirk
"""

import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
    TestCase,
)
import varwg
from varwg.time_series_analysis import rain_stats


class Test(TestCase):
    def setUp(self):
        self.sequences = 10
        self.slen = 3
        self.n = self.sequences * self.slen
        rain_ = np.zeros(self.n).reshape(self.sequences, self.slen)
        rain_[::2] = 1
        self.rain = rain_.ravel()

    def tearDown(self):
        pass

    def test_trans_prob(self):
        pp = rain_stats.trans_prob(self.rain)
        p11 = p00 = (self.slen - 1.0) * self.sequences / (2 * self.n)
        p10 = self.sequences / (2.0 * self.n)
        p01 = (self.sequences / 2.0 - 1) / self.n
        pp_exp = np.array([[p00, p01], [p10, p11]])
        assert_almost_equal(pp_exp, pp)

    def test_spells(self):
        dry, wet = rain_stats.spell_lengths(self.rain)
        expected = np.full(int(self.sequences / 2), 3)
        assert_equal(dry, expected)
        assert_equal(wet, expected)

    def test_richardson_model(self):
        varwg.reseed(0)
        # trans_probs = np.array([[0.33020467, 0.16190849],
        #                         [0.16177813, 0.34597836]])
        # trans_probs = np.array([[0.330, 0.162],
        #                         [0.162, 0.346]])
        trans_probs = np.array([[0.613, 0.145], [0.145, 0.096]])
        occurrences = rain_stats.richardson_model_occ(1e6, trans_probs)
        emp_trans_probs = rain_stats.trans_prob(occurrences)
        assert_almost_equal(trans_probs, emp_trans_probs, 3)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    # run_module_suite()
    pass



"""Central configuration file for all things VG.
This dummy-file contains necessary variables used as configurations and
suggestions for their values in the form of comments

Adjust this file for your needs and rename it to "config.py".
"""

import os
import tempfile

# -----------------------------------------------------------------------------#
# coordinates of measurement station
# they are used to determine the maximum possible short-wave radiation
# the following values point to constance, germany
latitude = 47.66
longitude = 9.18
# -----------------------------------------------------------------------------#
# by defining this variable we can set other directories relative to it (see
# data_dir)
script_home = os.path.dirname(__file__)
# data_dir = os.path.join(script_home, "MyProject", "data")
# data_dir = "i am just a dummy"
data_dir = os.path.join(tempfile.gettempdir(), "VG")

# the met_file can also be specified when initializing the VG object
# i.e. 'my_vg = varwg.VG(var_names, met_file="mymetfile")'
# met_file = "i am just a dummy"
met_file = os.path.join(script_home, "sample.met")
# -----------------------------------------------------------------------------#
cache_dir = os.path.join(data_dir, "cache")
out_dir = data_dir
seasonal_cache_file = os.path.join(cache_dir, "seasonal_solutions.she")
# this is a very experimental feature
# hyd_file = os.path.join(data_dir, "rhein_und_br_a_interp_jd.txt")
# -----------------------------------------------------------------------------#
# further suggestion: uncomment this to ensure the existence of the needed
# directories
# for directory in (data_dir, out_dir, cache_dir):
#    if (directory != "i am just a dummy") and (not os.path.exists(directory)):
#        os.makedirs(directory)


# -----------------------------------------------------------------------------#
# Configuration of how we call things
# -----------------------------------------------------------------------------#

import collections

long_var_names = {
    "R": "Precipitation",
    "U": "Wind velocity",
    "u": "Eastward wind speed",
    "v": "Northward wind speed",
    "ILWR": "Incident long-wave radiation",
    "Qsw": "Short-wave radiation",
    "Qsw_lars": "Short-wave radiation",
    "Cloud_Cover": "Cloud cover",
    "theta": "Temperature",
    "theta_lars": "Temperature",
    "theta_monthly": "Monthly averaged Temperature",
    "e": "Vapour pressure",
    "rh": "Relative humidity",
    "nao": "North Atlantic Oscillation Index",
    "wtemp_bregenzerach": "Water temperature of Bregenzer Ach",
    "wtemp_rhein": "Water temperature of Rhine",
}
units = collections.defaultdict(lambda: r"-")
units.update(
    {
        "ILWR": "[$W / m ^ 2$]",
        "e": "[hPa]",
        "rh": "[-]",
        "Cloud_Cover": "[-]",
        "R": "[mm]",
        "Qsw": "[$W / m ^ 2$]",
        "Qsw_lars": "[$W / m ^ 2$]",
        "theta": "[$^{\circ}C$]",
        "theta_lars": "[$^{\circ}C$]",
        "theta_monthly": "[$^{\circ}C$]",
        "U": "[$m / s$]",
        "u": "[$m / s$]",
        "v": "[$m / s$]",
        "nao": "?",
    }
)
# long variable names together with units
ylabels = [
    "%s %s" % (long_var_names[var_name], units[var_name])
    for var_name in long_var_names
]
ygreek = collections.defaultdict(lambda: r"-")
ygreek.update(
    {
        "R": r"$R$",
        "theta": r"$\theta$",
        "rh": r"$\phi$",
        "Qsw": r"$Q_{sw}$",
        "ILWR": r"$Q_{lw(inc.)}$",
        "u": r"$u$",
        "v": r"$v$",
    }
)


def var_names_greek(var_names):
    return [ygreek[var_name] for var_name in var_names]


# -----------------------------------------------------------------------------#
# These are internals of the fitting / deseasonalization process
# -----------------------------------------------------------------------------#

import numpy as np
from varwg.time_series_analysis import distributions
from varwg.time_series_analysis import seasonal_distributions as sd
from varwg.time_series_analysis import seasonal_kde as skde
from varwg.meteo import meteox2y
from varwg import helpers as my

# theoretical distributions for each variable.
dists = {
    "ILWR": distributions.norm,
    "e": distributions.norm,
    "rh": distributions.truncnorm,
    "Cloud_Cover": distributions.truncnorm,
    "Qsw": "empirical",
    "theta": distributions.norm,
    "R": (distributions.RainMix, distributions.kumaraswamy),
    "U": distributions.lognormal,
    "u": "empirical",
    "v": "empirical",
    "nao": distributions.norm,
}
seasonal_classes = dict.fromkeys(dists, sd.SlidingDist)
seasonal_classes["Qsw"] = skde.SeasonalKDE
seasonal_classes["u"] = seasonal_classes["v"] = skde.SeasonalKDE

# threshold for daily precipitation sum to count as such. same unit as
# precipitation in the input file
threshold = 0.0005
dists_kwds = {
    "R": dict(
        threshold=threshold,
        q_thresh_lower=0.7,
        q_thresh_upper=0.99,
        doy_width=20,
        fft_order=4,
        tabulate_cdf=True,
    ),
    # "R": dict(threshold=1.5e-4,
    #           doy_width=35, fft_order=4),
    "rh": dict(doy_width=10),
}


# this is a helper function for the definition of par_known
# functions that return known parameters to fix those values during fitting
def array_gen(scalar):
    return lambda tt: np.full_like(tt, scalar)


# uncomment the following decorator to cache results of max_qsw on the
# hard-drive and speed up repeated vg initialization
# @my.pickle_cache(os.path.join(cache_dir, "max_qsw_%s.pkl"), warn=False)
def max_qsw(doys, lon=longitude, lat=latitude, **kwds):
    """Use meteo.meteox2y to sum up hourly values of solar radiation for the
    specified days - of - year."""
    doys = np.atleast_1d(doys)
    qsw_2d = -1 * np.ones((22 - 4, len(doys)))
    for i, doy in enumerate(doys):
        date = [str("%03d %02d" % (doy, hour)) for hour in range(4, 22)]
        qsw_2d[:, i] = meteox2y.pot_s_rad(
            date, longt=lon, lat=lat, in_format="%j %H", **kwds
        )
    return qsw_2d.sum(axis=0).ravel()


# read: hours per day
hpd = 24
par_known = dict.fromkeys(dists)
par_known.update(
    {
        "Cloud_Cover": {"l": array_gen(-0.1), "u": array_gen(hpd + 0.1)},
        "Qsw": {
            "l": array_gen(0),
            # the .725 is unfortunately a complicated
            # subject it seems as it should also be a
            # function of t
            "u": lambda doys: 0.725 * np.array([
                max_qsw(int(d + 1 if d < 366 else 1))[0] for d in np.atleast_1d(doys)
            ]),
        },
        "rh": {
            "lc": array_gen(-0.001),
            "uc": array_gen(hpd * 1.01),
            # "l": array_gen(-.001),
            # "u": array_gen(hpd * 1.01)
        },
        "R": {
            "l": array_gen(0.0),
            "u": array_gen(1.0),
            # "kernel_bounds": array_gen([0.1, None]),
            # "lc": array_gen(.0002),
            # "uc": array_gen(1e12)
        },
    }
)


# some of the variables have to be treated differently in the hourly
# discretization
seasonal_classes_hourly = {key: val for key, val in seasonal_classes.items()}
# seasonal_classes_hourly["rh"] = skde.SeasonalHourlyKDE
dists_hourly = {key: val for key, val in list(dists.items())}
dists_hourly.update({"rh": distributions.Censored(distributions.norm)})

par_known_hourly = {key: val for key, val in list(par_known.items())}
par_known_hourly.update(
    {
        "rh": {
            "lc": array_gen(-0.001),
            "uc": array_gen(1.01),
        },
        "R": {
            "uc": array_gen(15.0),
            # "lc": array_gen(old_div(dists_kwds["R"]["threshold"], hpd)),
        },
        "Qsw": {
            "lc": array_gen(0),
            # the .725 is unfortunately a complicated subject it seems
            # as it should also be a function of t
            "uc": lambda doy: 0.725 * meteox2y.pot_s_rad(doy),
            # "uc": np.vectorize(lambda x: hpd ** -1 * .725 *
            #                    max_qsw(x + 1 if x < 366
            #                            else 1))
        },
    }
)

# -----------------------------------------------------------------------------#
# Here you can define conversions done after the simulation and before
# output of the data.
# The functions should accept a datetime array an array with the data and
# an iterable containing the variable names.
# -----------------------------------------------------------------------------#

conversions = []


def R_m2mm(times_, data, var_names, inverse=False):
    try:
        ri = var_names.index("R")
    except ValueError:
        return times_, data, var_names
    if inverse:
        data[ri] /= 1000 * 24
    else:
        data[ri] *= 1000 * 24
    return times_, data, var_names


conversions += [R_m2mm]

# # for example: calculate long-wave radiation from air temperature
# def theta2ilwr(times_, data, var_names):
#     from scipy import constants
#     theta = data[var_names.index("theta")]
#     theta_kelvin = constants.C2K(theta)
#     e = 6.2 * np.exp(17.26 * theta / (theta_kelvin - 35.8))
#     boltz = 0.0000000567  # ask gideon
#     alpha, beta = .42, .065
#     ilwr = boltz * theta_kelvin ** 4 * (alpha + beta * np.sqrt(e))
#     if "ILWR" not in var_names:
#         data = np.vstack((data, ilwr))
#         var_names = tuple(list(var_names) + ["ILWR"])
#     else:
#         data[var_names.index("ILWR")] = ilwr
#     return times_, data, var_names
#
# conversions += [theta2ilwr]


# -----------------------------------------------------------------------------#
# Define the format the output should have
# -----------------------------------------------------------------------------#

# if not specified differently, output will have 3 decimal places
out_format = collections.defaultdict(lambda: "%.3f")
out_format.update(
    {
        "theta": "%.3f",
        "Qsw": "%.3f",
        "ILWR": "%.3f",
        "rh": "%.3f",
        "U": "%.3f",
        "wdir": "%.1f",
        "u": "%.3f",
        "v": "%.3f",
        "R": "%.6f",
    }
)
